import pytest

from uaclient import messages
from uaclient.entitlements.entitlement_status import (
    CanDisableFailure,
    CanDisableFailureReason,
)


class TestCanDisableFailure:
    @pytest.mark.parametrize(
        "can_disable_failure,expected",
        (
            (
                CanDisableFailure(
                    reason=CanDisableFailureReason.ALREADY_DISABLED,
                    message=None,
                ),
                "",
            ),
            (
                CanDisableFailure(
                    reason=CanDisableFailureReason.NOT_APPLICABLE,
                    message=messages.NamedMessage("test", "test"),
                ),
                "test",
            ),
        ),
    )
    def test_message_value(self, can_disable_failure, expected):
        assert expected == can_disable_failure.message_value



"""
Client version related functions
"""

import os.path
from math import inf
from typing import Optional

from uaclient.apt import (
    get_apt_cache_time,
    get_pkg_candidate_version,
    version_compare,
)
from uaclient.defaults import CANDIDATE_CACHE_PATH, UAC_RUN_PATH
from uaclient.exceptions import ProcessExecutionError
from uaclient.system import subp

__VERSION__ = "37.2"
PACKAGED_VERSION = "@@PACKAGED_VERSION@@"


def get_version() -> str:
    """Return the packaged version as a string

    Prefer the binary PACKAGED_VESION set by debian/rules to DEB_VERSION.
    If unavailable, check for a .git development environments:
      a. If run in our upstream repo `git describe` will gives a leading
         XX.Y so return the --long version to allow daily build recipes
         to count commit offset from upstream's XX.Y signed tag.
      b. If run in a git-ubuntu pkg repo, upstream tags aren't visible,
         believe __VERSION__ is correct - there is and MUST always be a
         test to make sure it matches debian/changelog
    """
    if not PACKAGED_VERSION.startswith("@@PACKAGED_VERSION"):
        return PACKAGED_VERSION
    topdir = os.path.dirname(os.path.dirname(__file__))
    if os.path.exists(os.path.join(topdir, ".git")):
        cmd = ["git", "describe", "--abbrev=8", "--match=[0-9]*", "--long"]
        try:
            out, _ = subp(cmd)
            return out.strip()
        except ProcessExecutionError:
            pass
    return __VERSION__


def get_last_known_candidate() -> Optional[str]:
    # If we can't determine when the cache was updated for the last time,
    # We always assume it was as recent as possible - thus `inf`.
    last_apt_cache_update = get_apt_cache_time() or inf
    if (
        not os.path.exists(CANDIDATE_CACHE_PATH)
        or os.stat(CANDIDATE_CACHE_PATH).st_mtime < last_apt_cache_update
    ):
        candidate_version = None
        try:
            candidate_version = get_pkg_candidate_version("ubuntu-pro-client")
            if candidate_version:
                os.makedirs(UAC_RUN_PATH, exist_ok=True)
                with open(CANDIDATE_CACHE_PATH, "w") as f:
                    f.write(candidate_version)
                return candidate_version
        except Exception:
            if candidate_version is not None:
                return candidate_version

    try:
        with open(CANDIDATE_CACHE_PATH, "r") as f:
            return f.read().strip()
    except Exception:
        pass

    return None


def check_for_new_version() -> Optional[str]:
    candidate = get_last_known_candidate()
    if candidate and version_compare(candidate, get_version()) > 0:
        return candidate
    return None



import logging
import os
import time
from typing import Tuple

from uaclient import exceptions, system, util
from uaclient.data_types import DataObject, Field, StringDataValue
from uaclient.files import notices
from uaclient.files.data_types import DataObjectFile, DataObjectFileFormat
from uaclient.files.files import UAFile
from uaclient.files.notices import Notice

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


class LockData(DataObject):
    fields = [
        Field("lock_pid", StringDataValue),
        Field("lock_holder", StringDataValue),
    ]

    def __init__(self, lock_pid: str, lock_holder: str):
        self.lock_pid = lock_pid
        self.lock_holder = lock_holder


lock_data_file = DataObjectFile(
    LockData,
    UAFile("lock", private=False),
    DataObjectFileFormat.JSON,
)


def check_lock_info() -> Tuple[int, str]:
    """Return lock info if lock file is present the lock is active.

    If process claiming the lock is no longer present, remove the lock file
    and log a warning.

    :return: A tuple (pid, string describing lock holder)
        If no active lock, pid will be -1.
    """

    try:
        lock_data_obj = lock_data_file.read()
    except exceptions.InvalidFileFormatError:
        raise exceptions.InvalidLockFile(lock_file_path=lock_data_file.path)

    no_lock = (-1, "")
    if not lock_data_obj:
        return no_lock

    lock_pid = lock_data_obj.lock_pid
    lock_holder = lock_data_obj.lock_holder

    try:
        system.subp(["ps", lock_pid])
        return (int(lock_pid), lock_holder)
    except exceptions.ProcessExecutionError:
        if not util.we_are_currently_root():
            LOG.debug(
                "Found stale lock file previously held by %s:%s",
                lock_pid,
                lock_holder,
            )
            return (int(lock_pid), lock_holder)
        LOG.warning(
            "Removing stale lock file previously held by %s:%s",
            lock_pid,
            lock_holder,
        )
        system.ensure_file_absent(lock_data_file.path)
        return no_lock


def clear_lock_file_if_present():
    lock_data_file.delete()


class RetryLock:
    """
    Context manager for gaining exclusive access to the lock file.

    Create a lock file if absent. The lock file will contain a pid of the
    running process, and a customer-visible description of the lock holder.

    The RetryLock will try several times to acquire the lock before giving up.
    The number of times to try and how long to sleep in between tries is
    configurable.

    :param lock_holder: String with the service name or command which is
        holding the lock. This lock_holder string will be customer visible in
        status.json.
    :param sleep_time: Number of seconds to sleep before retrying if the lock
        is already held.
    :param max_retries: Maximum number of times to try to grab the lock before
        giving up and raising a LockHeldError.
    :raises: LockHeldError if lock is held after (sleep_time * max_retries)
    """

    def __init__(
        self,
        *_args,
        lock_holder: str,
        sleep_time: int = 10,
        max_retries: int = 12
    ):
        self.lock_holder = lock_holder
        self.sleep_time = sleep_time
        self.max_retries = max_retries

    def grab_lock(self):
        (lock_pid, cur_lock_holder) = check_lock_info()
        if lock_pid > 0:
            raise exceptions.LockHeldError(
                lock_request=self.lock_holder,
                lock_holder=cur_lock_holder,
                pid=lock_pid,
            )
        lock_data_file.write(
            LockData(lock_pid=str(os.getpid()), lock_holder=self.lock_holder)
        )
        notices.add(
            Notice.OPERATION_IN_PROGRESS,
            operation=self.lock_holder,
        )

    def __enter__(self):
        LOG.debug("spin lock starting for %s", self.lock_holder)
        tries = 0
        while True:
            try:
                self.grab_lock()
                break
            except exceptions.LockHeldError as e:
                LOG.debug(
                    "RetryLock Attempt %d. %s. Spinning...", tries + 1, e.msg
                )
                tries += 1
                if tries >= self.max_retries:
                    raise e
                else:
                    time.sleep(self.sleep_time)

    def __exit__(self, _exc_type, _exc_value, _traceback):
        lock_data_file.delete()
        notices.remove(Notice.OPERATION_IN_PROGRESS)



#!/usr/bin/env python3

"""
Perform auto-attach operation

On Ubuntu Pro machines, we try to perform an auto-attach operation
on first boot. This happens through a systemd unit that executes
this script on every boot. However, if we detect
that cloud-init has user data related to ua, we don't run
auto-attach here, since cloud-init will drive this operation on
their side.
"""
import logging
import sys

from uaclient import exceptions, http, log, messages, system
from uaclient.api.exceptions import (
    AlreadyAttachedError,
    AutoAttachDisabledError,
    EntitlementsNotEnabledError,
)
from uaclient.api.u.pro.attach.auto.full_auto_attach.v1 import (
    FullAutoAttachOptions,
    full_auto_attach,
)
from uaclient.clouds.aws import AWSAutoAttachInstance
from uaclient.clouds.identity import cloud_instance_factory
from uaclient.config import UAConfig
from uaclient.daemon import AUTO_ATTACH_STATUS_MOTD_FILE, retry_auto_attach
from uaclient.files import state_files

LOG = logging.getLogger("ubuntupro.lib.auto_attach")

# All known cloud-config keys which provide ubuntu pro configuration directives
CLOUD_INIT_UA_KEYS = set(
    ["ubuntu-advantage", "ubuntu_advantage", "ubuntu_pro"]
)

try:
    import cloudinit.stages as ci_stages  # type: ignore
except ImportError:
    pass


def get_cloudinit_init_stage():
    if "cloudinit.stages" in sys.modules:
        return ci_stages.Init()

    return None


def check_cloudinit_userdata_for_ua_info():
    init = get_cloudinit_init_stage()

    # if init is None, this means we were not able to import the cloud-init
    # module.
    if init is None:
        return False

    if init.cfg and CLOUD_INIT_UA_KEYS.intersection(init.cfg):
        return True

    return False


def main(cfg: UAConfig):
    if check_cloudinit_userdata_for_ua_info():
        LOG.info("cloud-init userdata has ubuntu-advantage key.")
        LOG.info(
            "Skipping auto-attach and deferring to cloud-init "
            "to setup and configure auto-attach"
        )
        return

    try:
        cloud = cloud_instance_factory()
    except exceptions.CloudFactoryError as e:
        LOG.debug("Error loading the cloud: %s", e)
    else:
        if isinstance(cloud, AWSAutoAttachInstance):
            if not cloud.is_likely_pro:
                LOG.info(
                    "Skipping auto-attach. Reason: No billingProduct nor"
                    " marketplaceProductCode on AWS."
                )
                return
            else:
                LOG.info("Auto-attaching: product code found on AWS.")

    system.write_file(
        AUTO_ATTACH_STATUS_MOTD_FILE, messages.AUTO_ATTACH_RUNNING
    )
    try:
        full_auto_attach(FullAutoAttachOptions())
    except AlreadyAttachedError as e:
        LOG.info(e.msg)
    except AutoAttachDisabledError:
        LOG.debug("Skipping auto-attach. Config disable_auto_attach is set.")
    except EntitlementsNotEnabledError as e:
        LOG.warning(e.msg)
    except Exception as e:
        LOG.error(e)
        system.ensure_file_absent(AUTO_ATTACH_STATUS_MOTD_FILE)
        LOG.info("creating flag file to trigger retries")
        system.create_file(retry_auto_attach.FLAG_FILE_PATH)
        failure_reason = (
            retry_auto_attach.full_auto_attach_exception_to_failure_reason(e)
        )
        state_files.retry_auto_attach_state_file.write(
            state_files.RetryAutoAttachState(
                interval_index=0, failure_reason=failure_reason
            )
        )
        return 1

    system.ensure_file_absent(AUTO_ATTACH_STATUS_MOTD_FILE)
    return 0


if __name__ == "__main__":
    log.setup_journald_logging()
    cfg = UAConfig()
    http.configure_web_proxy(cfg.http_proxy, cfg.https_proxy)
    sys.exit(main(cfg))



import logging
from subprocess import TimeoutExpired

from uaclient import exceptions, system, util

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


def start():
    try:
        system.subp(["systemctl", "start", "ua-timer.timer"], timeout=2.0)
    except (exceptions.ProcessExecutionError, TimeoutExpired) as e:
        LOG.warning(e, exc_info=e)


def stop():
    try:
        system.subp(["systemctl", "stop", "ua-timer.timer"], timeout=2.0)
    except (exceptions.ProcessExecutionError, TimeoutExpired) as e:
        LOG.warning(e, exc_info=e)



from uaclient import messages
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.api.errors import APIError
from uaclient.config import UAConfig
from uaclient.data_types import DataObject, Field, StringDataValue
from uaclient.version import get_version


class VersionError(APIError):
    _formatted_msg = messages.E_API_VERSION_ERROR


class VersionResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "installed_version",
            StringDataValue,
            doc="The current installed version",
        ),
    ]

    def __init__(self, *, installed_version: str):
        self.installed_version = installed_version


def version() -> VersionResult:
    return _version(UAConfig())


def _version(cfg: UAConfig) -> VersionResult:
    """
    This endpoint shows the installed Pro Client version.
    """
    try:
        version = get_version()
    except Exception as e:
        raise VersionError(error_msg=str(e))
    return VersionResult(installed_version=version)


endpoint = APIEndpoint(
    version="v1",
    name="Version",
    fn=_version,
    options_cls=None,
)

_doc = {
    "introduced_in": "27.11",
    "requires_network": False,
    "example_python": """
from uaclient.api.u.pro.version.v1 import version

result = version()
""",
    "result_class": VersionResult,
    "exceptions": [
        (VersionError, "Raised if the Client cannot determine the version.")
    ],
    "example_cli": "pro api u.pro.version.v1",
    "example_json": """
{
    "installed_version": "32.3~24.04"
}
""",
}



import json
import logging

import click

from features.util import build_debs, repo_state_hash


@click.command()
@click.argument("series")
@click.option(
    "-c",
    "--chroot",
    type=str,
    help="Use the supplied chroot when calling sbuild",
)
@click.option(
    "-a",
    "--arch",
    type=str,
    help="Architecture to use in the build",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    default=False,
    help="Don't show sbuild output",
)
def main(series=None, chroot=None, arch=None, quiet=False):
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)
    print(
        json.dumps(
            {
                "state_hash": repo_state_hash(),
                "debs": build_debs(
                    series,
                    chroot=chroot,
                    sbuild_output_to_terminal=not quiet,
                    architecture=arch,
                ).__dict__,
            }
        )
    )


if __name__ == "__main__":
    main()



import copy
import logging
import socket
from collections import namedtuple
from typing import Any, Dict, List, Optional, Tuple

import uaclient.files.machine_token as mtf
from uaclient import (
    data_types,
    event_logger,
    exceptions,
    http,
    messages,
    secret_manager,
    system,
    util,
    version,
)
from uaclient.api.u.pro.status.enabled_services.v1 import _enabled_services
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.config import UAConfig
from uaclient.defaults import ATTACH_FAIL_DATE_FORMAT
from uaclient.files.state_files import attachment_data_file, machine_id_file
from uaclient.http import serviceclient
from uaclient.log import get_user_or_root_log_file_path

# Here we describe every endpoint from the ua-contracts
# service that is used by this client implementation.
API_V1_ADD_CONTRACT_MACHINE = "/v1/context/machines/token"
API_V1_GET_CONTRACT_MACHINE = (
    "/v1/contracts/{contract}/context/machines/{machine}"
)
API_V1_UPDATE_CONTRACT_MACHINE = (
    "/v1/contracts/{contract}/context/machines/{machine}"
)
API_V1_AVAILABLE_RESOURCES = "/v1/resources"
API_V1_GET_RESOURCE_MACHINE_ACCESS = (
    "/v1/resources/{resource}/context/machines/{machine}"
)
API_V1_GET_CONTRACT_TOKEN_FOR_CLOUD_INSTANCE = "/v1/clouds/{cloud_type}/token"
API_V1_UPDATE_ACTIVITY_TOKEN = (
    "/v1/contracts/{contract}/machine-activity/{machine}"
)
API_V1_GET_CONTRACT_USING_TOKEN = "/v1/contract"

API_V1_GET_MAGIC_ATTACH_TOKEN_INFO = "/v1/magic-attach"
API_V1_NEW_MAGIC_ATTACH = "/v1/magic-attach"
API_V1_REVOKE_MAGIC_ATTACH = "/v1/magic-attach"

API_V1_GET_GUEST_TOKEN = (
    "/v1/contracts/{contract}/context/machines/{machine}/guest-token"
)

OVERRIDE_SELECTOR_WEIGHTS = {
    "series_overrides": 1,
    "series": 2,
    "cloud": 3,
    "variant": 4,
}

event = event_logger.get_event_logger()
LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))


EnableByDefaultService = namedtuple(
    "EnableByDefaultService", ["name", "variant"]
)


class CPUTypeData(data_types.DataObject):
    fields = [
        data_types.Field(
            "cpuinfo_cpu", data_types.StringDataValue, required=False
        ),
        data_types.Field(
            "cpuinfo_cpu_architecture",
            data_types.StringDataValue,
            required=False,
        ),
        data_types.Field(
            "cpuinfo_cpu_family", data_types.StringDataValue, required=False
        ),
        data_types.Field(
            "cpuinfo_cpu_implementer",
            data_types.StringDataValue,
            required=False,
        ),
        data_types.Field(
            "cpuinfo_cpu_part", data_types.StringDataValue, required=False
        ),
        data_types.Field(
            "cpuinfo_cpu_revision", data_types.StringDataValue, required=False
        ),
        data_types.Field(
            "cpuinfo_cpu_variant", data_types.StringDataValue, required=False
        ),
        data_types.Field(
            "cpuinfo_model", data_types.StringDataValue, required=False
        ),
        data_types.Field(
            "cpuinfo_model_name", data_types.StringDataValue, required=False
        ),
        data_types.Field(
            "cpuinfo_stepping", data_types.StringDataValue, required=False
        ),
        data_types.Field(
            "cpuinfo_vendor_id", data_types.StringDataValue, required=False
        ),
        data_types.Field(
            "sys_firmware_devicetree_base_model",
            data_types.StringDataValue,
            required=False,
        ),
        data_types.Field(
            "sysinfo_model", data_types.StringDataValue, required=False
        ),
        data_types.Field(
            "sysinfo_type", data_types.StringDataValue, required=False
        ),
    ]

    def __init__(
        self,
        cpuinfo_cpu: Optional[str] = None,
        cpuinfo_cpu_architecture: Optional[str] = None,
        cpuinfo_cpu_family: Optional[str] = None,
        cpuinfo_cpu_implementer: Optional[str] = None,
        cpuinfo_cpu_part: Optional[str] = None,
        cpuinfo_cpu_revision: Optional[str] = None,
        cpuinfo_cpu_variant: Optional[str] = None,
        cpuinfo_model: Optional[str] = None,
        cpuinfo_model_name: Optional[str] = None,
        cpuinfo_stepping: Optional[str] = None,
        cpuinfo_vendor_id: Optional[str] = None,
        sys_firmware_devicetree_base_model: Optional[str] = None,
        sysinfo_model: Optional[str] = None,
        sysinfo_type: Optional[str] = None,
    ):
        self.cpuinfo_cpu = cpuinfo_cpu
        self.cpuinfo_cpu_architecture = cpuinfo_cpu_architecture
        self.cpuinfo_cpu_family = cpuinfo_cpu_family
        self.cpuinfo_cpu_implementer = cpuinfo_cpu_implementer
        self.cpuinfo_cpu_part = cpuinfo_cpu_part
        self.cpuinfo_cpu_revision = cpuinfo_cpu_revision
        self.cpuinfo_cpu_variant = cpuinfo_cpu_variant
        self.cpuinfo_model = cpuinfo_model
        self.cpuinfo_model_name = cpuinfo_model_name
        self.cpuinfo_stepping = cpuinfo_stepping
        self.cpuinfo_vendor_id = cpuinfo_vendor_id
        self.sys_firmware_devicetree_base_model = (
            sys_firmware_devicetree_base_model
        )
        self.sysinfo_model = sysinfo_model
        self.sysinfo_type = sysinfo_type


class UAContractClient(serviceclient.UAServiceClient):
    cfg_url_base_attr = "contract_url"

    def __init__(
        self,
        cfg: Optional[UAConfig] = None,
    ) -> None:
        super().__init__(cfg=cfg)
        self.machine_token_file = mtf.get_machine_token_file()

    @util.retry(socket.timeout, retry_sleeps=[1, 2, 2])
    def add_contract_machine(
        self, contract_token, attachment_dt, machine_id=None
    ):
        """Requests machine attach to the provided machine_id.

        @param contract_token: Token string providing authentication to
            ContractBearer service endpoint.
        @param machine_id: Optional unique system machine id. When absent,
            contents of /etc/machine-id will be used.

        @return: Dict of the JSON response containing the machine-token.
        """
        if not machine_id:
            machine_id = system.get_machine_id(self.cfg)

        headers = self.headers()
        headers.update({"Authorization": "Bearer {}".format(contract_token)})
        activity_info = self._get_activity_info()
        activity_info["lastAttachment"] = attachment_dt.isoformat()
        data = {"machineId": machine_id, "activityInfo": activity_info}
        backcompat_data = _support_old_machine_info(data)
        response = self.request_url(
            API_V1_ADD_CONTRACT_MACHINE, data=backcompat_data, headers=headers
        )
        if response.code == 401:
            raise exceptions.AttachInvalidTokenError()
        elif response.code == 403:
            _raise_attach_forbidden_message(response)
        if response.code != 200:
            raise exceptions.ContractAPIError(
                url=API_V1_ADD_CONTRACT_MACHINE,
                code=response.code,
                body=response.body,
            )
        response_json = response.json_dict
        secret_manager.secrets.add_secret(
            response_json.get("machineToken", "")
        )
        for token in response_json.get("resourceTokens", []):
            secret_manager.secrets.add_secret(token.get("token", ""))
        return response_json

    def available_resources(self) -> Dict[str, Any]:
        """Requests list of entitlements available to this machine type."""
        activity_info = self._get_activity_info()
        response = self.request_url(
            API_V1_AVAILABLE_RESOURCES,
            query_params={
                "architecture": activity_info["architecture"],
                "series": activity_info["series"],
                "kernel": activity_info["kernel"],
                "virt": activity_info["virt"],
            },
        )
        if response.code != 200:
            raise exceptions.ContractAPIError(
                url=API_V1_AVAILABLE_RESOURCES,
                code=response.code,
                body=response.body,
            )
        return response.json_dict

    def get_contract_using_token(self, contract_token: str) -> Dict[str, Any]:
        headers = self.headers()
        headers.update({"Authorization": "Bearer {}".format(contract_token)})
        response = self.request_url(
            API_V1_GET_CONTRACT_USING_TOKEN, headers=headers
        )
        if response.code != 200:
            raise exceptions.ContractAPIError(
                url=API_V1_GET_CONTRACT_USING_TOKEN,
                code=response.code,
                body=response.body,
            )
        return response.json_dict

    @util.retry(socket.timeout, retry_sleeps=[1, 2, 2])
    def get_contract_token_for_cloud_instance(
        self, *, cloud_type: str, data: Dict[str, Any]
    ):
        """Requests contract token for auto-attach images for Pro clouds.

        @param instance: AutoAttachCloudInstance for the cloud.

        @return: Dict of the JSON response containing the contract-token.
        """
        response = self.request_url(
            API_V1_GET_CONTRACT_TOKEN_FOR_CLOUD_INSTANCE.format(
                cloud_type=cloud_type
            ),
            data=data,
        )
        if response.code != 200:
            msg = response.json_dict.get("message", "")
            if msg:
                LOG.debug(msg)
                raise exceptions.InvalidProImage(error_msg=msg)
            raise exceptions.ContractAPIError(
                url=API_V1_GET_CONTRACT_TOKEN_FOR_CLOUD_INSTANCE,
                code=response.code,
                body=response.body,
            )

        response_json = response.json_dict
        secret_manager.secrets.add_secret(
            response_json.get("contractToken", "")
        )
        return response_json

    def get_resource_machine_access(
        self,
        machine_token: str,
        resource: str,
        machine_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Requests machine access context for a given resource

        @param machine_token: The authentication token needed to talk to
            this contract service endpoint.
        @param resource: Entitlement name.
        @param machine_id: Optional unique system machine id. When absent,
            contents of /etc/machine-id will be used.

        @return: Dict of the JSON response containing entitlement accessInfo.
        """
        if not machine_id:
            machine_id = system.get_machine_id(self.cfg)
        headers = self.headers()
        headers.update({"Authorization": "Bearer {}".format(machine_token)})
        url = API_V1_GET_RESOURCE_MACHINE_ACCESS.format(
            resource=resource, machine=machine_id
        )
        response = self.request_url(url, headers=headers)
        if response.code != 200:
            raise exceptions.ContractAPIError(
                url=API_V1_GET_RESOURCE_MACHINE_ACCESS,
                code=response.code,
                body=response.body,
            )
        if response.headers.get("expires"):
            response.json_dict["expires"] = response.headers["expires"]

        response_json = response.json_dict
        for token in response_json.get("resourceTokens", []):
            secret_manager.secrets.add_secret(token.get("token", ""))
        return response_json

    def update_activity_token(self):
        """Report current activity token and enabled services.

        This will report to the contracts backend all the current
        enabled services in the system.
        """
        contract_id = self.machine_token_file.contract_id
        machine_token = self.machine_token_file.machine_token.get(
            "machineToken"
        )
        machine_id = system.get_machine_id(self.cfg)

        request_data = self._get_activity_info()
        url = API_V1_UPDATE_ACTIVITY_TOKEN.format(
            contract=contract_id, machine=machine_id
        )
        headers = self.headers()
        headers.update({"Authorization": "Bearer {}".format(machine_token)})

        response = self.request_url(url, headers=headers, data=request_data)
        if response.code != 200:
            raise exceptions.ContractAPIError(
                url=url, code=response.code, body=response.body
            )

        # We will update the `machine-token.json` based on the response
        # provided by the server. We expect the response to be
        # a full `activityInfo` object which belongs at the root of
        # `machine-token.json`
        if response.json_dict:
            machine_token = self.machine_token_file.machine_token
            # The activity information received as a response here
            # will not provide the information inside an activityInfo
            # structure. However, this structure will be reflected when
            # we reach the contract for attach and refresh requests.
            # Because of that, we will store the response directly on
            # the activityInfo key
            machine_token["activityInfo"] = response.json_dict
            self.machine_token_file.write(machine_token)

    def get_magic_attach_token_info(self, magic_token: str) -> Dict[str, Any]:
        """Request magic attach token info.

        When the magic token is registered, it will contain new fields
        that will allow us to know that the attach process can proceed
        """
        headers = self.headers()
        headers.update({"Authorization": "Bearer {}".format(magic_token)})
        response = self.request_url(
            API_V1_GET_MAGIC_ATTACH_TOKEN_INFO, headers=headers
        )

        if response.code == 401:
            raise exceptions.MagicAttachTokenError()
        if response.code == 503:
            raise exceptions.MagicAttachUnavailable()
        if response.code != 200:
            raise exceptions.ContractAPIError(
                url=API_V1_GET_MAGIC_ATTACH_TOKEN_INFO,
                code=response.code,
                body=response.body,
            )
        response_json = response.json_dict
        secret_fields = ["token", "userCode", "contractToken"]
        for field in secret_fields:
            secret_manager.secrets.add_secret(response_json.get(field, ""))
        return response_json

    def new_magic_attach_token(self) -> Dict[str, Any]:
        """Create a magic attach token for the user."""
        headers = self.headers()
        response = self.request_url(
            API_V1_NEW_MAGIC_ATTACH,
            headers=headers,
            method="POST",
        )

        if response.code == 503:
            raise exceptions.MagicAttachUnavailable()
        if response.code != 200:
            raise exceptions.ContractAPIError(
                url=API_V1_NEW_MAGIC_ATTACH,
                code=response.code,
                body=response.body,
            )
        response_json = response.json_dict
        secret_fields = ["token", "userCode", "contractToken"]
        for field in secret_fields:
            secret_manager.secrets.add_secret(response_json.get(field, ""))
        return response_json

    def revoke_magic_attach_token(self, magic_token: str):
        """Revoke a magic attach token for the user."""
        headers = self.headers()
        headers.update({"Authorization": "Bearer {}".format(magic_token)})
        response = self.request_url(
            API_V1_REVOKE_MAGIC_ATTACH,
            headers=headers,
            method="DELETE",
        )

        if response.code == 400:
            raise exceptions.MagicAttachTokenAlreadyActivated()
        if response.code == 401:
            raise exceptions.MagicAttachTokenError()
        if response.code == 503:
            raise exceptions.MagicAttachUnavailable()
        if response.code != 200:
            raise exceptions.ContractAPIError(
                url=API_V1_REVOKE_MAGIC_ATTACH,
                code=response.code,
                body=response.body,
            )

    def get_contract_machine(
        self,
        machine_token: str,
        contract_id: str,
        machine_id: Optional[str] = None,
    ) -> Dict[str, Any]:
        """Get the updated machine token from the contract server.

        @param machine_token: The machine token needed to talk to
            this contract service endpoint.
        @param contract_id: Unique contract id provided by contract service
        @param machine_id: Optional unique system machine id. When absent,
            contents of /etc/machine-id will be used.
        """
        if not machine_id:
            machine_id = system.get_machine_id(self.cfg)

        headers = self.headers()
        headers.update({"Authorization": "Bearer {}".format(machine_token)})
        url = API_V1_GET_CONTRACT_MACHINE.format(
            contract=contract_id,
            machine=machine_id,
        )
        activity_info = self._get_activity_info()
        response = self.request_url(
            url,
            method="GET",
            headers=headers,
            query_params={
                "architecture": activity_info["architecture"],
                "series": activity_info["series"],
                "kernel": activity_info["kernel"],
                "virt": activity_info["virt"],
            },
        )
        if response.code != 200:
            raise exceptions.ContractAPIError(
                url=url, code=response.code, body=response.body
            )
        if response.headers.get("expires"):
            response.json_dict["expires"] = response.headers["expires"]
        return response.json_dict

    def update_contract_machine(
        self,
        machine_token: str,
        contract_id: str,
        machine_id: Optional[str] = None,
    ) -> Dict:
        """Request machine token refresh from contract server.

        @param machine_token: The machine token needed to talk to
            this contract service endpoint.
        @param contract_id: Unique contract id provided by contract service.
        @param machine_id: Optional unique system machine id. When absent,
            contents of /etc/machine-id will be used.

        @return: Dict of the JSON response containing refreshed machine-token
        """
        if not machine_id:
            machine_id = system.get_machine_id(self.cfg)

        headers = self.headers()
        headers.update({"Authorization": "Bearer {}".format(machine_token)})
        data = {
            "machineId": machine_id,
            "activityInfo": self._get_activity_info(),
        }
        backcompat_data = _support_old_machine_info(data)
        url = API_V1_UPDATE_CONTRACT_MACHINE.format(
            contract=contract_id, machine=machine_id
        )
        response = self.request_url(
            url, headers=headers, method="POST", data=backcompat_data
        )
        if response.code != 200:
            raise exceptions.ContractAPIError(
                url=url, code=response.code, body=response.body
            )
        if response.headers.get("expires"):
            response.json_dict["expires"] = response.headers["expires"]
        return response.json_dict

    def get_guest_token(
        self,
        machine_token: str,
        contract_id: str,
        machine_id: str,
    ) -> Dict:
        """Request guest token associated with this machine's contract
        @param machine_token: The machine token needed to talk to
            this contract service endpoint.
        @param contract_id: Unique contract id provided by contract service
        @param machine_id: Unique machine id that was registered with the pro
            backend on attach.
        @return: Dict of the JSON response containing the guest token
        """
        headers = self.headers()
        headers.update({"Authorization": "Bearer {}".format(machine_token)})
        url = API_V1_GET_GUEST_TOKEN.format(
            contract=contract_id,
            machine=machine_id,
        )
        response = self.request_url(url, headers=headers, method="GET")
        if response.code == 400:
            # defined response code for when machine has a v1 or v2 token
            # this will only be true for old attachments (2020 or earlier)
            raise exceptions.FeatureNotSupportedOldTokenError(
                feature_name="get_guest_token"
            )
        elif response.code != 200:
            raise exceptions.ContractAPIError(
                url=url,
                code=response.code,
                body=response.body,
            )
        return response.json_dict

    def _get_activity_info(self):
        """Return a dict of activity info data for contract requests"""

        cpuinfo = system.get_cpu_info()
        machine_info = {
            "distribution": system.get_release_info().distribution,
            "kernel": system.get_kernel_info().uname_release,
            "series": system.get_release_info().series,
            "architecture": system.get_dpkg_arch(),
            "desktop": system.is_desktop(),
            "virt": system.get_virt_type(),
            "clientVersion": version.get_version(),
            "cpu_type": CPUTypeData(
                cpuinfo_cpu=cpuinfo.cpuinfo_cpu,
                cpuinfo_cpu_architecture=cpuinfo.cpuinfo_cpu_architecture,
                cpuinfo_cpu_family=cpuinfo.cpuinfo_cpu_family,
                cpuinfo_cpu_implementer=cpuinfo.cpuinfo_cpu_implementer,
                cpuinfo_cpu_part=cpuinfo.cpuinfo_cpu_part,
                cpuinfo_cpu_revision=cpuinfo.cpuinfo_cpu_revision,
                cpuinfo_cpu_variant=cpuinfo.cpuinfo_cpu_variant,
                cpuinfo_model=cpuinfo.cpuinfo_model,
                cpuinfo_model_name=cpuinfo.cpuinfo_model_name,
                cpuinfo_stepping=cpuinfo.cpuinfo_stepping,
                cpuinfo_vendor_id=cpuinfo.cpuinfo_vendor_id,
                sys_firmware_devicetree_base_model=cpuinfo.sys_firmware_devicetree_base_model,  # noqa: E501
                sysinfo_model=cpuinfo.sysinfo_model,
                sysinfo_type=cpuinfo.sysinfo_type,
            ).to_dict(keep_none=False),
        }

        if _is_attached(self.cfg).is_attached:
            enabled_services = _enabled_services(self.cfg).enabled_services
            attachment_data = attachment_data_file.read()
            activity_info = {
                "activityID": self.machine_token_file.activity_id
                or system.get_machine_id(self.cfg),
                "activityToken": self.machine_token_file.activity_token,
                "resources": [service.name for service in enabled_services],
                "resourceVariants": {
                    service.name: service.variant_name
                    for service in enabled_services
                    if service.variant_enabled
                },
                "lastAttachment": (
                    attachment_data.attached_at.isoformat()
                    if attachment_data
                    else None
                ),
            }
        else:
            activity_info = {}

        return {
            **activity_info,
            **machine_info,
        }


def _support_old_machine_info(request_body: dict):
    """
    Transforms a request_body that has the new activity_info into a body that
    includes both old and new forms of machineInfo/activityInfo

    This is necessary because there may be old ua-airgapped contract
    servers deployed that we need to support.
    This function is used for attach and refresh calls.
    """
    activity_info = request_body.get("activityInfo", {})

    return {
        "machineId": request_body.get("machineId"),
        "activityInfo": activity_info,
        "architecture": activity_info.get("architecture"),
        "os": {
            "distribution": activity_info.get("distribution"),
            "kernel": activity_info.get("kernel"),
            "series": activity_info.get("series"),
            # These two are required for old ua-airgapped but not sent anymore
            # in the new activityInfo
            "type": "Linux",
            "release": system.get_release_info().release,
        },
    }


def process_entitlements_delta(
    cfg: UAConfig,
    past_entitlements: Dict[str, Any],
    new_entitlements: Dict[str, Any],
    allow_enable: bool,
    series_overrides: bool = True,
    verbose: bool = True,
) -> None:
    """Iterate over all entitlements in new_entitlement and apply any delta
    found according to past_entitlements.

    :param cfg: UAConfig instance
    :param past_entitlements: dict containing the last valid information
        regarding service entitlements.
    :param new_entitlements: dict containing the current information regarding
        service entitlements.
    :param allow_enable: Boolean set True if allowed to perform the enable
        operation. When False, a message will be logged to inform the user
        about the recommended enabled service.
    :param series_overrides: Boolean set True if series overrides should be
        applied to the new_access dict.
    :param verbose: If True, display output to stdout
    """
    from uaclient.entitlements import entitlements_enable_order

    delta_error = False
    unexpected_errors = []

    # We need to sort our entitlements because some of them
    # depend on other service to be enable first.
    failed_services = []  # type: List[str]
    for name in entitlements_enable_order(cfg):
        try:
            new_entitlement = new_entitlements[name]
        except KeyError:
            continue

        failed_services = []
        try:
            deltas, service_enabled = process_entitlement_delta(
                cfg=cfg,
                orig_access=past_entitlements.get(name, {}),
                new_access=new_entitlement,
                allow_enable=allow_enable,
                series_overrides=series_overrides,
                verbose=verbose,
            )
        except exceptions.UbuntuProError as e:
            LOG.exception(e)
            delta_error = True
            failed_services.append(name)
            LOG.error(
                "Failed to process contract delta for %s: %r",
                name,
                new_entitlement,
            )
        except Exception as e:
            LOG.exception(e)
            unexpected_errors.append(e)
            failed_services.append(name)
            LOG.exception(
                "Unexpected error processing contract delta for %s: %r",
                name,
                new_entitlement,
            )
        else:
            # If we have any deltas to process and we were able to process
            # them, then we will mark that service as successfully enabled
            if service_enabled and deltas:
                event.service_processed(name)
    event.services_failed(failed_services)
    if len(unexpected_errors) > 0:
        raise exceptions.AttachFailureUnknownError(
            failed_services=[
                (
                    name,
                    messages.UNEXPECTED_ERROR.format(
                        error_msg=str(exception),
                        log_path=get_user_or_root_log_file_path(),
                    ),
                )
                for name, exception in zip(failed_services, unexpected_errors)
            ]
        )
    elif delta_error:
        raise exceptions.AttachFailureDefaultServices(
            failed_services=[
                (name, messages.E_ATTACH_FAILURE_DEFAULT_SERVICES)
                for name in failed_services
            ]
        )


def process_entitlement_delta(
    cfg: UAConfig,
    orig_access: Dict[str, Any],
    new_access: Dict[str, Any],
    allow_enable: bool = False,
    series_overrides: bool = True,
    verbose: bool = True,
) -> Tuple[Dict, bool]:
    """Process a entitlement access dictionary deltas if they exist.

    :param cfg: UAConfig instance
    :param orig_access: Dict with original entitlement access details before
        contract refresh deltas
    :param new_access: Dict with updated entitlement access details after
        contract refresh
    :param allow_enable: Boolean set True if allowed to perform the enable
        operation. When False, a message will be logged to inform the user
        about the recommended enabled service.
    :param series_overrides: Boolean set True if series overrides should be
        applied to the new_access dict.
    :param verbose: If True, display output to stdout

    :raise UbuntuProError: on failure to process deltas.
    :return: A tuple containing a dict of processed deltas and a
             boolean indicating if the service was fully processed
    """
    from uaclient.entitlements import entitlement_factory

    if series_overrides:
        apply_contract_overrides(new_access)

    deltas = util.get_dict_deltas(orig_access, new_access)
    ret = False
    if deltas:
        name = orig_access.get("entitlement", {}).get("type")
        if not name:
            name = deltas.get("entitlement", {}).get("type")
        if not name:
            raise exceptions.InvalidContractDeltasServiceType(
                orig=orig_access, new=new_access
            )

        variant = (
            new_access.get("entitlements", {})
            .get("obligations", {})
            .get("use_selector", "")
        )
        try:
            entitlement = entitlement_factory(
                cfg=cfg,
                name=name,
                variant=variant,
            )
        except exceptions.EntitlementNotFoundError as exc:
            LOG.debug(
                'Skipping entitlement deltas for "%s". No such class', name
            )
            raise exc

        ret = entitlement.process_contract_deltas(
            orig_access, deltas, allow_enable=allow_enable, verbose=verbose
        )
    return deltas, ret


def _raise_attach_forbidden_message(
    response: http.HTTPResponse,
) -> messages.NamedMessage:
    info = response.json_dict.get("info")
    if info:
        contract_id = info["contractId"]
        reason = info["reason"]
        if reason == "no-longer-effective":
            date = info["time"].strftime(ATTACH_FAIL_DATE_FORMAT)
            raise exceptions.AttachForbiddenExpired(
                contract_id=contract_id,
                date=date,
                # keep this extra date for backwards compat
                contract_expiry_date=info["time"].strftime("%m-%d-%Y"),
            )
        elif reason == "not-effective-yet":
            date = info["time"].strftime(ATTACH_FAIL_DATE_FORMAT)
            raise exceptions.AttachForbiddenNotYet(
                contract_id=contract_id,
                date=date,
                # keep this extra date for backwards compat
                contract_effective_date=info["time"].strftime("%m-%d-%Y"),
            )
        elif reason == "never-effective":
            raise exceptions.AttachForbiddenNever(contract_id=contract_id)

    raise exceptions.AttachExpiredToken()


def refresh(cfg: UAConfig, verbose=True):
    """Request contract refresh from ua-contracts service.

    :param verbose: If True, display output to stdout

    :raise UbuntuProError: on failure to update contract or error processing
        contract deltas
    :raise ConnectivityError: On failure during a connection
    """
    machine_token_file = mtf.get_machine_token_file(cfg)
    orig_entitlements = machine_token_file.entitlements()
    orig_token = machine_token_file.machine_token
    machine_token = orig_token["machineToken"]
    contract_id = orig_token["machineTokenInfo"]["contractInfo"]["id"]

    contract_client = UAContractClient(cfg=cfg)
    resp = contract_client.update_contract_machine(
        machine_token=machine_token, contract_id=contract_id
    )

    machine_token_file.write(resp)
    system.get_machine_id.cache_clear()
    machine_id = resp.get("machineTokenInfo", {}).get(
        "machineId", system.get_machine_id(cfg)
    )
    machine_id_file.write(machine_id)

    process_entitlements_delta(
        cfg,
        orig_entitlements,
        machine_token_file.entitlements(),
        allow_enable=False,
        verbose=verbose,
    )


def get_available_resources(cfg: UAConfig) -> List[Dict]:
    """Query available resources from the contract server for this machine."""
    client = UAContractClient(cfg)
    resources = client.available_resources()
    return resources.get("resources", [])


def get_contract_information(cfg: UAConfig, token: str) -> Dict[str, Any]:
    """Query contract information for a specific token"""
    client = UAContractClient(cfg)
    return client.get_contract_using_token(token)


def _get_override_weight(
    override_selector: Dict[str, str], selector_values: Dict[str, str]
) -> int:
    override_weight = 0
    for selector, value in override_selector.items():
        if (selector, value) not in selector_values.items():
            return 0
        override_weight += OVERRIDE_SELECTOR_WEIGHTS[selector]

    return override_weight


def _select_overrides(
    entitlement: Dict[str, Any],
    series_name: str,
    cloud_type: str,
    variant: Optional[str] = None,
) -> Dict[int, Dict[str, Any]]:
    overrides = {}

    selector_values = {"series": series_name, "cloud": cloud_type}
    if variant:
        selector_values["variant"] = variant

    series_overrides = entitlement.pop("series", {}).pop(series_name, {})
    if series_overrides:
        overrides[OVERRIDE_SELECTOR_WEIGHTS["series_overrides"]] = (
            series_overrides
        )

    general_overrides = copy.deepcopy(entitlement.get("overrides", []))
    for override in general_overrides:
        weight = _get_override_weight(
            override.pop("selector"), selector_values
        )
        if weight:
            overrides[weight] = override

    return overrides


def apply_contract_overrides(
    orig_access: Dict[str, Any],
    series: Optional[str] = None,
    variant: Optional[str] = None,
) -> None:
    """Apply series-specific overrides to an entitlement dict.

    This function mutates orig_access dict by applying any series-overrides to
    the top-level keys under 'entitlement'. The series-overrides are sparse
    and intended to supplement existing top-level dict values. So, sub-keys
    under the top-level directives, obligations and affordance sub-key values
    will be preserved if unspecified in series-overrides.

    To more clearly indicate that orig_access in memory has already had
    the overrides applied, the 'series' key is also removed from the
    orig_access dict.

    :param orig_access: Dict with original entitlement access details
    """
    from uaclient.clouds.identity import get_cloud_type

    if not all([isinstance(orig_access, dict), "entitlement" in orig_access]):
        raise RuntimeError(
            'Expected entitlement access dict. Missing "entitlement" key:'
            " {}".format(orig_access)
        )

    series_name = (
        system.get_release_info().series if series is None else series
    )
    cloud_type, _ = get_cloud_type()
    orig_entitlement = orig_access.get("entitlement", {})

    overrides = _select_overrides(
        orig_entitlement, series_name, cloud_type, variant
    )

    for _weight, overrides_to_apply in sorted(overrides.items()):
        for key, value in overrides_to_apply.items():
            current = orig_access["entitlement"].get(key)
            if isinstance(current, dict):
                # If the key already exists and is a dict,
                # update that dict using the override
                current.update(value)
            else:
                # Otherwise, replace it wholesale
                orig_access["entitlement"][key] = value


def get_enabled_by_default_services(
    cfg: UAConfig, entitlements: Dict[str, Any]
) -> List[EnableByDefaultService]:
    from uaclient.entitlements import entitlement_factory

    enable_by_default_services = []

    for ent_name, ent_value in entitlements.items():
        variant = ent_value.get("obligations", {}).get("use_selector", "")

        try:
            ent = entitlement_factory(cfg=cfg, name=ent_name, variant=variant)
        except exceptions.EntitlementNotFoundError:
            continue

        obligations = ent_value.get("entitlement", {}).get("obligations", {})
        resourceToken = ent_value.get("resourceToken")

        if ent._should_enable_by_default(obligations, resourceToken):
            can_enable, _ = ent.can_enable()
            if can_enable:
                enable_by_default_services.append(
                    EnableByDefaultService(
                        name=ent_name,
                        variant=variant,
                    )
                )

    return enable_by_default_services



from typing import Any, Dict, List, Union  # noqa: F401

from uaclient.data_types import DataObject, Field, StringDataValue, data_list
from uaclient.util import get_pro_environment
from uaclient.version import get_version


class AdditionalInfo:
    meta = {}  # type: Dict[str, Any]
    warnings = []  # type: List[ErrorWarningObject]


class ErrorWarningObject(DataObject):
    fields = [
        Field("title", StringDataValue),
        Field("code", StringDataValue),
        Field("meta", DataObject),
    ]

    def __init__(self, *, title: str, code: str, meta: dict):
        self.title = title
        self.code = code
        self.meta = meta


class APIData(DataObject):
    fields = [
        Field("type", StringDataValue),
        Field("attributes", DataObject),
        Field("meta", DataObject),
    ]

    def __init__(self, *, type: str, attributes: DataObject, meta: dict):
        self.type = type
        self.attributes = attributes
        self.meta = {
            "environment_vars": [
                {"name": name, "value": value}
                for name, value in sorted(get_pro_environment().items())
            ],
            **meta,
        }


class APIResponse(DataObject):
    fields = [
        Field("_schema_version", StringDataValue),
        Field("result", StringDataValue),
        Field("version", StringDataValue),
        Field("errors", data_list(ErrorWarningObject)),
        Field("warnings", data_list(ErrorWarningObject)),
        Field("data", APIData),
    ]

    def __init__(
        self,
        *,
        _schema_version: str,
        result: str = "success",
        errors: List[ErrorWarningObject] = [],
        warnings: List[ErrorWarningObject] = [],
        data: Union[APIData, dict]
    ):
        self._schema_version = _schema_version
        self.result = result
        self.version = get_version()
        self.errors = errors
        self.warnings = warnings
        self.data = data



import copy
import logging
import os
from functools import lru_cache, wraps
from typing import Any, Dict, Optional

from uaclient import (
    apt,
    event_logger,
    exceptions,
    http,
    messages,
    snap,
    system,
    util,
)
from uaclient.defaults import (
    APT_NEWS_URL,
    BASE_CONTRACT_URL,
    BASE_LIVEPATCH_URL,
    BASE_SECURITY_URL,
    BASE_VULNERABILITY_DATA_URL,
    CONFIG_DEFAULTS,
    CONFIG_FIELD_ENVVAR_ALLOWLIST,
    DEFAULT_CONFIG_FILE,
    DEFAULT_DATA_DIR,
)
from uaclient.files import state_files, user_config_file
from uaclient.yaml import safe_load

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))

MERGE_ID_KEY_MAP = {
    "availableResources": "name",
    "resourceEntitlements": "type",
}
UNSET_SETTINGS_OVERRIDE_KEY = "_unset"

# Keys visible and configurable using `pro config set|unset|show` subcommands
UA_CONFIGURABLE_KEYS = (
    "http_proxy",
    "https_proxy",
    "apt_http_proxy",
    "apt_https_proxy",
    "ua_apt_http_proxy",
    "ua_apt_https_proxy",
    "global_apt_http_proxy",
    "global_apt_https_proxy",
    "update_messaging_timer",
    "metering_timer",
    "apt_news",
    "apt_news_url",
    "vulnerability_data_url_prefix",
    "lxd_guest_attach",
)

# Basic schema validation top-level keys for parse_config handling
VALID_UA_CONFIG_KEYS = (
    "contract_url",
    "data_dir",
    "features",
    "log_file",
    "log_level",
    "security_url",
    "vulnerability_data_url_prefix",
    "settings_overrides",
    "ua_config",
    "livepatch_url",
)


event = event_logger.get_event_logger()


class UAConfig:
    ua_scoped_proxy_options = ("ua_apt_http_proxy", "ua_apt_https_proxy")
    global_scoped_proxy_options = (
        "global_apt_http_proxy",
        "global_apt_https_proxy",
    )
    deprecated_global_scoped_proxy_options = (
        "apt_http_proxy",
        "apt_https_proxy",
    )

    def __init__(
        self,
        cfg: Optional[Dict[str, Any]] = None,
        user_config: Optional[user_config_file.UserConfigData] = None,
        series: Optional[str] = None,
    ) -> None:
        """"""
        if cfg:
            self.cfg_path = None
            self.cfg = cfg
            self.invalid_keys = None
        else:
            self.cfg_path = get_config_path()
            self.cfg, self.invalid_keys = parse_config(self.cfg_path)

        if user_config:
            self.user_config = user_config
        else:
            try:
                self.user_config = user_config_file.user_config.read()
            except Exception as e:
                LOG.warning("Error loading user config", exc_info=e)
                LOG.warning("Using default config values")
                self.user_config = user_config_file.UserConfigData()

        # support old ua_config values in uaclient.conf as user-config.json
        # value overrides
        if "ua_config" in self.cfg:
            self.user_config = user_config_file.UserConfigData.from_dict(
                {**self.user_config.to_dict(), **self.cfg["ua_config"]},
                optional_type_errors_become_null=True,
            )

        self.series = series

    @property
    def contract_url(self) -> str:
        return self.cfg.get("contract_url", BASE_CONTRACT_URL)

    @property
    def security_url(self) -> str:
        return self.cfg.get("security_url", BASE_SECURITY_URL)

    @property
    def livepatch_url(self) -> str:
        return self.cfg.get("livepatch_url", BASE_LIVEPATCH_URL)

    @property
    def http_proxy(self) -> Optional[str]:
        return self.user_config.http_proxy

    @http_proxy.setter
    def http_proxy(self, value: str):
        self.user_config.http_proxy = value
        user_config_file.user_config.write(self.user_config)

    @property
    def https_proxy(self) -> Optional[str]:
        return self.user_config.https_proxy

    @https_proxy.setter
    def https_proxy(self, value: str):
        self.user_config.https_proxy = value
        user_config_file.user_config.write(self.user_config)

    @property
    def ua_apt_https_proxy(self) -> Optional[str]:
        return self.user_config.ua_apt_https_proxy

    @ua_apt_https_proxy.setter
    def ua_apt_https_proxy(self, value: str):
        self.user_config.ua_apt_https_proxy = value
        user_config_file.user_config.write(self.user_config)

    @property
    def ua_apt_http_proxy(self) -> Optional[str]:
        return self.user_config.ua_apt_http_proxy

    @ua_apt_http_proxy.setter
    def ua_apt_http_proxy(self, value: str):
        self.user_config.ua_apt_http_proxy = value
        user_config_file.user_config.write(self.user_config)

    @property
    def vulnerability_data_url_prefix(self) -> str:
        val = self.user_config.vulnerability_data_url_prefix
        if val is None:
            return BASE_VULNERABILITY_DATA_URL
        return val

    @vulnerability_data_url_prefix.setter
    def vulnerability_data_url_prefix(self, value: str):
        self.user_config.vulnerability_data_url_prefix = value
        user_config_file.user_config.write(self.user_config)

    @property  # type: ignore
    @lru_cache(maxsize=None)
    def global_apt_http_proxy(self) -> Optional[str]:
        global_val = self.user_config.global_apt_http_proxy
        if global_val:
            return global_val

        old_apt_val = self.user_config.apt_http_proxy
        if old_apt_val:
            event.info(
                messages.WARNING_DEPRECATED_FIELD.format(
                    old="apt_http_proxy", new="global_apt_http_proxy"
                )
            )
            return old_apt_val
        return None

    @global_apt_http_proxy.setter
    def global_apt_http_proxy(self, value: str):
        self.user_config.global_apt_http_proxy = value
        self.user_config.apt_http_proxy = None
        UAConfig.global_apt_http_proxy.fget.cache_clear()  # type: ignore
        user_config_file.user_config.write(self.user_config)

    @property  # type: ignore
    @lru_cache(maxsize=None)
    def global_apt_https_proxy(self) -> Optional[str]:
        global_val = self.user_config.global_apt_https_proxy
        if global_val:
            return global_val

        old_apt_val = self.user_config.apt_https_proxy
        if old_apt_val:
            event.info(
                messages.WARNING_DEPRECATED_FIELD.format(
                    old="apt_https_proxy", new="global_apt_https_proxy"
                )
            )
            return old_apt_val
        return None

    @global_apt_https_proxy.setter
    def global_apt_https_proxy(self, value: str):
        self.user_config.global_apt_https_proxy = value
        self.user_config.apt_https_proxy = None
        UAConfig.global_apt_https_proxy.fget.cache_clear()  # type: ignore
        user_config_file.user_config.write(self.user_config)

    @property
    def update_messaging_timer(self) -> int:
        val = self.user_config.update_messaging_timer
        if val is None:
            return 21600
        return val

    @update_messaging_timer.setter
    def update_messaging_timer(self, value: int):
        self.user_config.update_messaging_timer = value
        user_config_file.user_config.write(self.user_config)

    @property
    def metering_timer(self) -> int:
        val = self.user_config.metering_timer
        if val is None:
            return 14400
        return val

    @metering_timer.setter
    def metering_timer(self, value: int):
        self.user_config.metering_timer = value
        user_config_file.user_config.write(self.user_config)

    @property
    def poll_for_pro_license(self) -> bool:
        # TODO: when polling is supported
        #     1. change default here to True
        #     2. add this field to UA_CONFIGURABLE_KEYS
        val = self.user_config.poll_for_pro_license
        if val is None:
            return False
        return val

    @poll_for_pro_license.setter
    def poll_for_pro_license(self, value: bool):
        self.user_config.poll_for_pro_license = value
        user_config_file.user_config.write(self.user_config)

    @property
    def polling_error_retry_delay(self) -> int:
        # TODO: when polling is supported
        #     1. add this field to UA_CONFIGURABLE_KEYS
        val = self.user_config.polling_error_retry_delay
        if val is None:
            return 600
        return val

    @polling_error_retry_delay.setter
    def polling_error_retry_delay(self, value: int):
        self.user_config.polling_error_retry_delay = value
        user_config_file.user_config.write(self.user_config)

    @property
    def apt_news(self) -> bool:
        val = self.user_config.apt_news
        if val is None:
            return True
        return val

    @apt_news.setter
    def apt_news(self, value: bool):
        self.user_config.apt_news = value
        user_config_file.user_config.write(self.user_config)

    @property
    def apt_news_url(self) -> str:
        val = self.user_config.apt_news_url
        if val is None:
            return APT_NEWS_URL
        return val

    @apt_news_url.setter
    def apt_news_url(self, value: str):
        self.user_config.apt_news_url = value
        user_config_file.user_config.write(self.user_config)

    @property
    def lxd_guest_attach(self) -> user_config_file.LXDGuestAttachEnum:
        val = self.user_config.lxd_guest_attach
        if val is None:
            return user_config_file.LXDGuestAttachEnum.OFF
        return val

    @lxd_guest_attach.setter
    def lxd_guest_attach(self, value: user_config_file.LXDGuestAttachEnum):
        if value is None:
            value = user_config_file.LXDGuestAttachEnum.OFF
        self.user_config.lxd_guest_attach = value
        user_config_file.user_config.write(self.user_config)

    @property
    def data_dir(self):
        return self.cfg.get("data_dir", DEFAULT_DATA_DIR)

    @property
    def log_level(self):
        log_level = self.cfg.get("log_level", "DEBUG")
        try:
            return getattr(logging, log_level.upper())
        except AttributeError:
            return logging.DEBUG

    @property
    def log_file(self) -> str:
        return self.cfg.get("log_file", CONFIG_DEFAULTS["log_file"])

    @property
    def features(self):
        """Return a dictionary of any features provided in uaclient.conf."""
        features = self.cfg.get("features")
        if features:
            if isinstance(features, dict):
                return features
            else:
                LOG.warning(
                    "Unexpected uaclient.conf features value."
                    " Expected dict, but found %s",
                    features,
                )
        return {}

    def process_config(self):
        for prop in (
            "update_messaging_timer",
            "metering_timer",
        ):
            value = getattr(self, prop)
            if value is None:
                LOG.debug(
                    "No config set for %s, default value will be used.", prop
                )
            elif not isinstance(value, int) or value < 0:
                raise exceptions.InvalidPosIntConfigValue(
                    key=prop, value=value
                )

        if (self.global_apt_http_proxy or self.global_apt_https_proxy) and (
            self.ua_apt_http_proxy or self.ua_apt_https_proxy
        ):
            # Should we unset the config values?
            raise exceptions.InvalidProxyCombinationConfig()

        http.validate_proxy(
            "http",
            self.global_apt_http_proxy,
            http.PROXY_VALIDATION_APT_HTTP_URL,
        )
        http.validate_proxy(
            "https",
            self.global_apt_https_proxy,
            http.PROXY_VALIDATION_APT_HTTPS_URL,
        )
        http.validate_proxy(
            "http", self.ua_apt_http_proxy, http.PROXY_VALIDATION_APT_HTTP_URL
        )
        http.validate_proxy(
            "https",
            self.ua_apt_https_proxy,
            http.PROXY_VALIDATION_APT_HTTPS_URL,
        )
        http.validate_proxy(
            "http", self.http_proxy, http.PROXY_VALIDATION_SNAP_HTTP_URL
        )
        http.validate_proxy(
            "https", self.https_proxy, http.PROXY_VALIDATION_SNAP_HTTPS_URL
        )

        if self.global_apt_http_proxy or self.global_apt_https_proxy:
            apt.setup_apt_proxy(
                self.global_apt_http_proxy,
                self.global_apt_https_proxy,
                apt.AptProxyScope.GLOBAL,
            )
        elif self.ua_apt_http_proxy or self.ua_apt_https_proxy:
            apt.setup_apt_proxy(
                self.ua_apt_http_proxy,
                self.ua_apt_https_proxy,
                apt.AptProxyScope.UACLIENT,
            )

        services_with_proxies = []
        if snap.is_snapd_installed():
            snap.configure_snap_proxy(self.http_proxy, self.https_proxy)
            if (
                not self.http_proxy
                and snap.get_config_option_value(snap.HTTP_PROXY_OPTION)
            ) or (
                not self.https_proxy
                and snap.get_config_option_value(snap.HTTPS_PROXY_OPTION)
            ):
                services_with_proxies.append("snap")

        from uaclient import livepatch
        from uaclient.api.u.pro.status.enabled_services.v1 import (
            _enabled_services,
        )

        enabled_services = _enabled_services(self).enabled_services
        livepatch_enabled = any(
            ent for ent in enabled_services if ent.name == "livepatch"
        )

        if livepatch_enabled:
            livepatch.configure_livepatch_proxy(
                self.http_proxy, self.https_proxy
            )
            if (
                not self.http_proxy
                and livepatch.get_config_option_value(
                    livepatch.HTTP_PROXY_OPTION
                )
            ) or (
                not self.https_proxy
                and livepatch.get_config_option_value(
                    livepatch.HTTPS_PROXY_OPTION
                )
            ):
                services_with_proxies.append("livepatch")

        if len(services_with_proxies) > 0:
            services = ", ".join(services_with_proxies)
            print(
                messages.PROXY_DETECTED_BUT_NOT_CONFIGURED.format(
                    services=services
                )
            )

        from uaclient.api.u.pro.status.is_attached.v1 import _is_attached

        if (
            self.lxd_guest_attach != user_config_file.LXDGuestAttachEnum.OFF
            and not _is_attached(self).is_attached
        ):
            print(messages.WARNING_LXD_GUEST_ATTACH_SET_BUT_NOT_ATTACHED)
        else:
            state_files.lxd_pro_config_file.write(
                state_files.LXDProConfig(guest_attach=self.lxd_guest_attach)
            )

    def warn_about_invalid_keys(self):
        if self.invalid_keys is not None:
            for invalid_key in sorted(self.invalid_keys):
                LOG.warning(
                    "Ignoring invalid uaclient.conf key: %s", invalid_key
                )
        if "ua_config" in self.cfg:
            # this one is still technically supported but we want people to
            # migrate so it gets a special warning
            LOG.warning('legacy "ua_config" found in uaclient.conf')
            LOG.warning("Please do the following:")
            LOG.warning(
                "  1. run `sudo pro config set field=value` for each"
                ' field/value pair present under "ua_config" in'
                " /etc/ubuntu-advantage/uaclient.conf"
            )
            LOG.warning(
                '  2. Delete "ua_config" and all sub-fields in'
                " /etc/ubuntu-advantage/uaclient.conf"
            )


def get_config_path() -> str:
    """Get config path to be used when loading config dict."""
    config_file = os.environ.get("UA_CONFIG_FILE")
    if config_file:
        return config_file

    return DEFAULT_CONFIG_FILE


def parse_config(config_path=None):
    """Parse known Pro config file

    Attempt to find configuration in cwd and fallback to DEFAULT_CONFIG_FILE.
    Any missing configuration keys will be set to CONFIG_DEFAULTS.

    Values are overridden by any environment variable with prefix 'UA_'.

    @param config_path: Fullpath to pro configfile. If unspecified, use
        DEFAULT_CONFIG_FILE.

    @return: Dict of configuration values.
    """
    cfg = copy.copy(CONFIG_DEFAULTS)  # type: Dict[str, Any]

    if not config_path:
        config_path = get_config_path()

    LOG.debug("Using client configuration file at %s", config_path)
    if os.path.exists(config_path):
        cfg.update(safe_load(system.load_file(config_path)))
    env_keys = {}
    for key, value in os.environ.items():
        key = key.lower()
        if key.startswith("ua_"):
            # Strip leading UA_
            field_name = key[3:]
            if field_name.startswith("features_"):
                # Strip leading UA_FEATURES_
                feature_field_name = field_name[9:]

                # Users can provide a yaml file to override
                # config behavor. If they do, we are going
                # to load that yaml and update the config
                # with it
                if value.endswith("yaml"):
                    if os.path.exists(value):
                        value = safe_load(system.load_file(value))
                    else:
                        raise exceptions.InvalidFeatureYamlConfigValue(
                            filepath=value
                        )

                if "features" not in cfg:
                    cfg["features"] = {feature_field_name: value}
                else:
                    cfg["features"][feature_field_name] = value
            elif key in CONFIG_FIELD_ENVVAR_ALLOWLIST:
                env_keys[field_name] = value
    cfg.update(env_keys)
    if "data_dir" in cfg:
        cfg["data_dir"] = os.path.expanduser(cfg["data_dir"])
    for key in ("contract_url", "security_url"):
        if not http.is_service_url(cfg[key]):
            raise exceptions.InvalidURLConfigValue(key=key, value=cfg[key])

    invalid_keys = set(cfg.keys()).difference(VALID_UA_CONFIG_KEYS)
    for invalid_key in invalid_keys:
        cfg.pop(invalid_key)

    return cfg, invalid_keys


def apply_config_settings_override(override_key: str):
    """Decorator used to override function return by config settings.

    To identify if we should override the function return, we check
    if the config object has the expected override key, we use it
    has, we will use the key value as the function return. Otherwise
    we will call the function normally.

    @param override_key: key to be looked for in the settings_override
     entry in the config dict. If that key is present, we will return
     its value as the function return.
    """

    def wrapper(f):
        @wraps(f)
        def new_f():
            cfg, _ = parse_config()
            value_override = cfg.get("settings_overrides", {}).get(
                override_key, UNSET_SETTINGS_OVERRIDE_KEY
            )

            if value_override != UNSET_SETTINGS_OVERRIDE_KEY:
                if override_key == "cloud_type":
                    return (value_override, None)
                return value_override

            return f()

        return new_f

    return wrapper



import json
import re
import time

from behave import then, when
from hamcrest import assert_that, equal_to

from features.steps.files import when_i_create_file_with_content
from features.steps.packages import when_i_apt_install
from features.steps.shell import when_i_run_command, when_i_run_shell_command
from uaclient.defaults import DEFAULT_CONFIG_FILE
from uaclient.util import DatetimeAwareJSONDecoder

AUTOCOMPLETE_TEST_FILE = """\
#!/usr/bin/expect -f

set command [lindex $argv 0];
set send_human {.1 .3 1 .05 2}
set timeout 10

spawn /bin/bash
expect -exact "$"
send -h "$command \t\t"
expect -exact "$ $command"
"""


@when("I append the following on uaclient config:")
def when_i_append_to_uaclient_config(context):
    cmd = "printf '\n{}\n' > /tmp/uaclient.conf".format(context.text)
    cmd = 'sh -c "{}"'.format(cmd)
    when_i_run_command(context, cmd, "as non-root")

    cmd = "cat /tmp/uaclient.conf >> {}".format(DEFAULT_CONFIG_FILE)
    cmd = 'sh -c "{}"'.format(cmd)
    when_i_run_command(context, cmd, "with sudo")


@when("I wait `{seconds}` seconds")
def when_i_wait(context, seconds):
    time.sleep(int(seconds))


@then("I verify that the timer interval for `{job}` is `{interval}`")
def verify_timer_interval_for_job(context, job, interval):
    when_i_run_command(
        context, "cat /var/lib/ubuntu-advantage/jobs-status.json", "with sudo"
    )
    jobs_status = json.loads(
        context.process.stdout.strip(), cls=DatetimeAwareJSONDecoder
    )
    last_run = jobs_status[job]["last_run"]
    next_run = jobs_status[job]["next_run"]
    run_diff = next_run - last_run

    assert_that(run_diff.seconds, equal_to(int(interval)))


@when("I prepare the autocomplete test")
def prepare_autocomplete_test(context):
    when_i_run_command(context, "apt update", "with sudo")
    when_i_apt_install(context, "expect")
    when_i_create_file_with_content(
        context, "/tmp/test_autocomplete.exp", text=AUTOCOMPLETE_TEST_FILE
    )
    when_i_run_command(
        context, "chmod +x /tmp/test_autocomplete.exp", "with sudo"
    )


@when("I press tab twice to autocomplete the `{command}` command")
def when_i_autocomplete_command(context, command):
    when_i_run_command(
        context,
        "/tmp/test_autocomplete.exp '{}'".format(command),
        "as non-root",
    )


@then("I verify that the folder `{folder}` does not exist")
def then_folder_does_not_exist(context, folder):
    when_i_run_shell_command(
        context,
        command="ls {} || echo 'empty'".format(folder),
        user_spec="with sudo",
    )
    assert_that(context.process.stdout.strip(), equal_to("empty"))


@when("I regexify `{var_name}` stored var")
def regixify_stored_var(context, var_name):
    val = context.stored_vars.get(var_name)

    if val:
        context.stored_vars[var_name] = re.escape(val)



from typing import List, Optional

from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.api.u.pro.security.fix._common import (
    FixStatus,
    get_expected_overall_status,
)

# Some of these imports are intentionally not used in this module.
# The rationale is that we want users to import such Data Objects
# directly from the associated endpoints and not through the _common module
from uaclient.api.u.pro.security.fix._common.execute.v1 import (  # noqa: F401
    FailedUpgrade,
    FixExecuteError,
    FixExecuteResult,
    UpgradedPackage,
    _execute_fix,
)
from uaclient.api.u.pro.security.fix.usn.plan.v1 import (
    USNFixPlanOptions,
    _plan,
)
from uaclient.config import UAConfig
from uaclient.data_types import DataObject, Field, StringDataValue, data_list


class USNFixExecuteOptions(DataObject):
    fields = [
        Field(
            "usns",
            data_list(StringDataValue),
            doc="A list of USNs (i.e. USN-6188-1) titles",
        ),
    ]

    def __init__(self, usns: List[str]):
        self.usns = usns


class FixExecuteUSNResult(DataObject):
    fields = [
        Field(
            "target_usn",
            FixExecuteResult,
            doc="The ``FixExecuteResult`` for the target USN",
        ),
        Field(
            "related_usns",
            data_list(FixExecuteResult),
            required=False,
            doc="A list of ``FixExecuteResult`` objects for the related USNs",
        ),
    ]

    def __init__(
        self,
        target_usn: FixExecuteResult,
        related_usns: Optional[List[FixExecuteResult]] = None,
    ):
        self.target_usn = target_usn
        self.related_usns = related_usns


class USNAPIFixExecuteResult(DataObject):
    fields = [
        Field("status", StringDataValue, doc="The status of fixing the USNs"),
        Field(
            "usns",
            data_list(FixExecuteUSNResult),
            doc="A list of ``FixExecuteUSNResult`` objects",
        ),
    ]

    def __init__(self, status: str, usns: List[FixExecuteUSNResult]):
        self.status = status
        self.usns = usns


class USNSAPIFixExecuteResult(DataObject, AdditionalInfo):
    fields = [
        Field(
            "usns_data",
            USNAPIFixExecuteResult,
            doc="A list of ``USNAPIFixExecuteResult`` objects",
        )
    ]

    def __init__(self, usns_data: USNAPIFixExecuteResult):
        self.usns_data = usns_data


def execute(options: USNFixExecuteOptions) -> USNSAPIFixExecuteResult:
    return _execute(options, UAConfig())


def _execute(
    options: USNFixExecuteOptions, cfg: UAConfig
) -> USNSAPIFixExecuteResult:
    """
    This endpoint fixes the specified USNs on the machine.
    """
    fix_plan = _plan(USNFixPlanOptions(usns=options.usns), cfg=cfg)
    usns_result = []  # type: List[FixExecuteUSNResult]
    all_usns_status = FixStatus.SYSTEM_NOT_AFFECTED.value.msg

    for usn in fix_plan.usns_data.usns:
        target_usn_result = _execute_fix(usn.target_usn_plan)

        all_usns_status = get_expected_overall_status(
            all_usns_status, target_usn_result.status
        )

        usn_fix_execute_result = FixExecuteUSNResult(
            target_usn=target_usn_result
        )

        if (
            target_usn_result.status
            != FixStatus.SYSTEM_STILL_VULNERABLE.value.msg
        ):
            related_usns_result = []  # type: List[FixExecuteResult]

            for related_usn in usn.related_usns_plan:
                related_usns_result.append(_execute_fix(related_usn))

            usn_fix_execute_result.related_usns = related_usns_result

        usns_result.append(usn_fix_execute_result)

    return USNSAPIFixExecuteResult(
        usns_data=USNAPIFixExecuteResult(
            status=all_usns_status, usns=usns_result
        )
    )


endpoint = APIEndpoint(
    version="v1",
    name="USNFixExecute",
    fn=_execute,
    options_cls=USNFixExecuteOptions,
)

_doc = {
    "introduced_in": "30",
    "requires_network": True,
    "example_python": """
from uaclient.api.u.pro.security.fix.usn.execute.v1 import execute, USNFixExecuteOptions

options = USNFixExecuteOptions(usns=["USN-1234-1", "USN-1235-1"])
result = execute(options)
""",  # noqa: E501
    "result_class": USNSAPIFixExecuteResult,
    "exceptions": [],
    "example_cli": """pro api u.pro.security.fix.usn.execute.v1 --data '{"usns": ["USN-1234-1", "USN-1235-1"]}'""",  # noqa: E501
    "example_json": """
{
    "usns_data": {
        "status": "fixed",
        "usns": [
            {
                "target_usn": {
                    "title": "CVE-1234-56789",
                    "status": "fixed",
                    "upgraded_packages": {
                        "name": "pkg1",
                        "version": "1.1",
                        "pocket": "standard-updates"
                    },
                    "error": null
                },
                "related_usns": []
            }
        ]
    }
}
""",
    "extra": """
.. tab-item:: Explanation
    :sync: explanation

    When using the USN endpoint, the expected output is as follows:

    .. code-block:: json

        {
            "usns_data": {
                "status": "fixed",
                "usns": [
                    {
                        "target_usn": {
                            "title": "CVE-1234-56789",
                            "status": "fixed",
                            "upgraded_packages": {
                                "name": "pkg1",
                                "version": "1.1",
                                "pocket": "standard-updates"
                            },
                            "error": null
                        },
                        "related_usns": []
                    }
                ]
            }
        }

    From this output, we can see that the **usns_data** object contains two
    attributes:

    * **usns**: A list of USN objects detailing what happened during the fix
      operation.
    * **status**: The status of the fix operation considering **all** USNs.
      This means that if one USN cannot be fixed, this field will reflect that.
      Note that related USNs don't interfere with this attribute, meaning that
      a related USN can fail to be fixed without modifying the **status**
      value.

    Each **usn** object contains a reference for both **target_usn** and
    **related_usns**. The target is the USN requested to be fixed by the user,
    while related USNs are USNs that are related to the main USN and an
    attempt to fix them will be performed by the endpoint too. To better
    understand that distinction, please refer to
    :ref:`our explanation of CVEs and USNs <expl-cve-usn>`.

    With that said both **target_usn** object and any object from
    **related_usns** follow this structure:

    * **title**: The title of the USN.
    * **description**: The USN description.
    * **error**: Any error captured when fixing the USN will appear here. The
      error object will be detailed in a following section.
    * **status**: The expected status of the USN after the fix operation.
      There are three possible scenarios: **fixed**, **still-affected** and
      **not-affected**. The system is considered **still-affected** if there
      is something that prevents any required packages from being upgraded.
      The system is considered **not-affected** if the USN doesn't affect the
      system at all.
    * **upgraded_packages**: A list of UpgradedPackage objects referencing each
      package that was upgraded during the fix operation. The UpgradedPackage
      object always contain the **name** of the package, the **version** it was
      upgraded to and the **pocket** where the package upgrade came from.

    **What errors can be generated?**

    There some errors that can happen when executing this endpoint. For
    example, the system might require the user to attach to a Pro subscription
    to install the upgrades, or the user might run the command as non-root
    when a package upgrade is needed.

    In those situations, the error JSON error object will follow this
    representation:

    .. code-block:: json

        {
            "error_type": "error-type",
            "reason": "reason",
            "failed_upgrades": [
                {
                    "name": "pkg1",
                    "pocket": "esm-infra"
                }
            ]
        }

    We can see that the representation has the following fields:

    * **error_type**: The error type
    * **reason**: The explanation of why the error happened
    * **failed_upgrade**: A list of objects that always contain the name of the
      package that was not upgraded and the pocket where the upgrade would have
      come from.
""",
}



import mock
import pytest

from uaclient import entitlements
from uaclient.api import exceptions
from uaclient.api.u.pro.services.enable.v1 import (
    EnableOptions,
    EnableResult,
    _auto_select_variant,
    _enable,
)
from uaclient.testing.helpers import does_not_raise, mock_with_name_attr

M_PATH = "uaclient.api.u.pro.services.enable.v1."


class TestEnable:
    @pytest.mark.parametrize(
        [
            "options",
            "we_are_currently_root",
            "is_attached",
            "enabled_services_before",
            "enabled_services_names_after",
            "enable_result",
            "messaging_ops",
            "reboot_required",
            "expected_raises",
            "expected_result",
        ],
        [
            # not root
            (
                EnableOptions(service="s1"),
                False,
                False,
                None,
                None,
                None,
                None,
                None,
                pytest.raises(exceptions.NonRootUserError),
                None,
            ),
            # not attached
            (
                EnableOptions(service="s1"),
                True,
                False,
                None,
                None,
                None,
                None,
                None,
                pytest.raises(exceptions.UnattachedError),
                None,
            ),
            # landscape fail
            (
                EnableOptions(service="landscape"),
                True,
                True,
                None,
                None,
                None,
                None,
                None,
                pytest.raises(exceptions.NotSupported),
                None,
            ),
            # generic enable failure
            (
                EnableOptions(service="s1"),
                True,
                True,
                [],
                None,
                (False, None),
                None,
                None,
                pytest.raises(exceptions.EntitlementNotEnabledError),
                None,
            ),
            # success
            (
                EnableOptions(service="s1"),
                True,
                True,
                [],
                ["s1"],
                (True, None),
                {},
                False,
                does_not_raise(),
                EnableResult(
                    enabled=["s1"],
                    disabled=[],
                    reboot_required=False,
                    messages=[],
                ),
            ),
            # success already enabled no variant
            (
                EnableOptions(service="s1"),
                True,
                True,
                [
                    mock_with_name_attr(
                        name="s1", variant_enabled=False, variant_name=""
                    )
                ],
                None,
                None,
                None,
                None,
                does_not_raise(),
                EnableResult(
                    enabled=[],
                    disabled=[],
                    reboot_required=False,
                    messages=[],
                ),
            ),
            # success already enabled no variant specified but variant enabled
            (
                EnableOptions(service="s1"),
                True,
                True,
                [
                    mock_with_name_attr(
                        name="s1", variant_enabled=True, variant_name="v1"
                    )
                ],
                None,
                None,
                None,
                None,
                does_not_raise(),
                EnableResult(
                    enabled=[],
                    disabled=[],
                    reboot_required=False,
                    messages=[],
                ),
            ),
            # success already enabled variant specified same variant enabled
            (
                EnableOptions(service="s1", variant="v1"),
                True,
                True,
                [
                    mock_with_name_attr(
                        name="s1", variant_enabled=True, variant_name="v1"
                    )
                ],
                None,
                None,
                None,
                None,
                does_not_raise(),
                EnableResult(
                    enabled=[],
                    disabled=[],
                    reboot_required=False,
                    messages=[],
                ),
            ),
            # success with reboot required
            (
                EnableOptions(service="s1"),
                True,
                True,
                [],
                ["s1"],
                (True, None),
                {},
                True,
                does_not_raise(),
                EnableResult(
                    enabled=["s1"],
                    disabled=[],
                    reboot_required=True,
                    messages=[],
                ),
            ),
            # success with reboot required and different variant
            (
                EnableOptions(service="s1", variant="v1"),
                True,
                True,
                [
                    mock_with_name_attr(
                        name="s1", variant_enabled=True, variant_name="v2"
                    )
                ],
                ["s1"],
                (True, None),
                {},
                True,
                does_not_raise(),
                EnableResult(
                    enabled=[],
                    disabled=[],
                    reboot_required=True,
                    messages=[],
                ),
            ),
            # success with post enable messages
            (
                EnableOptions(service="s1"),
                True,
                True,
                [],
                ["s1"],
                (True, None),
                {"post_enable": ["one", "two", 3]},
                False,
                does_not_raise(),
                EnableResult(
                    enabled=["s1"],
                    disabled=[],
                    reboot_required=False,
                    messages=["one", "two"],
                ),
            ),
            # success with additional enablements and disablements
            (
                EnableOptions(service="s1"),
                True,
                True,
                [
                    mock_with_name_attr(
                        name="s2", variant_enabled=False, variant_name=""
                    )
                ],
                ["s1", "s3"],
                (True, None),
                {},
                False,
                does_not_raise(),
                EnableResult(
                    enabled=["s1", "s3"],
                    disabled=["s2"],
                    reboot_required=False,
                    messages=[],
                ),
            ),
        ],
    )
    @mock.patch(M_PATH + "status.status")
    @mock.patch(M_PATH + "lock.clear_lock_file_if_present")
    @mock.patch(M_PATH + "lock.RetryLock")
    @mock.patch(M_PATH + "_auto_select_variant")
    @mock.patch(M_PATH + "entitlements.entitlement_factory")
    @mock.patch(M_PATH + "_enabled_services_names")
    @mock.patch(M_PATH + "_enabled_services")
    @mock.patch(M_PATH + "_is_attached")
    @mock.patch(M_PATH + "util.we_are_currently_root")
    def test_enable(
        self,
        m_we_are_currently_root,
        m_is_attached,
        m_enabled_services,
        m_enabled_services_names,
        m_entitlement_factory,
        m_auto_select_variant,
        m_spin_lock,
        m_clear_lock_file_if_present,
        m_status,
        options,
        is_attached,
        we_are_currently_root,
        enabled_services_before,
        enabled_services_names_after,
        enable_result,
        messaging_ops,
        reboot_required,
        expected_raises,
        expected_result,
        FakeConfig,
    ):
        m_we_are_currently_root.return_value = we_are_currently_root
        m_is_attached.return_value = mock.MagicMock(is_attached=is_attached)
        m_enabled_services.return_value.enabled_services = (
            enabled_services_before
        )
        enabled_services_names_before = (
            [s.name for s in enabled_services_before]
            if enabled_services_before
            else enabled_services_before
        )
        m_enabled_services_names.side_effect = [
            enabled_services_names_before,
            enabled_services_names_after,
        ]
        m_ent = m_entitlement_factory.return_value
        m_auto_select_variant.return_value = (m_ent, None)
        m_ent.applicability_status.return_value = (
            entitlements.ApplicabilityStatus.APPLICABLE,
            None,
        )
        m_ent.enable.return_value = enable_result
        m_ent.messaging = messaging_ops
        m_ent._check_for_reboot.return_value = reboot_required

        cfg = FakeConfig()

        actual_result = None
        with expected_raises:
            actual_result = _enable(
                options, cfg, progress_object=mock.MagicMock()
            )

        assert actual_result == expected_result

        if expected_result is not None and len(expected_result.enabled) > 0:
            assert m_entitlement_factory.call_args_list == [
                mock.call(
                    cfg=cfg,
                    name=options.service,
                    variant=options.variant or "",
                    access_only=options.access_only,
                )
            ]
            assert m_ent.enable.call_args_list == [mock.call(mock.ANY)]

    @pytest.mark.parametrize(
        [
            "original_applicability_status",
            "original_is_variant",
            "original_variants",
            "auto_select_variant_return_value",
            "expected_original_enable_calls",
            "expected_variant_enable_calls",
        ],
        [
            # base entitlement not applicable
            (
                (entitlements.ApplicabilityStatus.INAPPLICABLE, None),
                True,
                {"variant": mock.MagicMock()},
                None,
                [mock.call(mock.ANY)],
                [],
            ),
            # not a variant
            (
                (entitlements.ApplicabilityStatus.APPLICABLE, None),
                False,
                {"variant": mock.MagicMock()},
                None,
                [mock.call(mock.ANY)],
                [],
            ),
            # is already a variant and therefore no variants
            (
                (entitlements.ApplicabilityStatus.APPLICABLE, None),
                True,
                {},
                None,
                [mock.call(mock.ANY)],
                [],
            ),
            # variants but nothing returned for some reason
            (
                (entitlements.ApplicabilityStatus.APPLICABLE, None),
                False,
                {"variant": mock.MagicMock()},
                None,
                [mock.call(mock.ANY)],
                [],
            ),
            # variant returned
            (
                (entitlements.ApplicabilityStatus.APPLICABLE, None),
                False,
                {"variant": mock.MagicMock()},
                (mock.MagicMock(), mock.MagicMock()),
                [],
                [mock.call(mock.ANY)],
            ),
        ],
    )
    @mock.patch(M_PATH + "status.status")
    @mock.patch(M_PATH + "lock.clear_lock_file_if_present")
    @mock.patch(M_PATH + "lock.RetryLock")
    @mock.patch(
        M_PATH + "_auto_select_variant",
    )
    @mock.patch(M_PATH + "entitlements.entitlement_factory")
    @mock.patch(M_PATH + "_enabled_services_names")
    @mock.patch(M_PATH + "_is_attached")
    @mock.patch(M_PATH + "util.we_are_currently_root")
    def test_enable_auto_selects_variant(
        self,
        m_we_are_currently_root,
        m_is_attached,
        m_enabled_services_names,
        m_entitlement_factory,
        m_auto_select_variant,
        m_spin_lock,
        m_clear_lock_file_if_present,
        m_status,
        original_applicability_status,
        original_is_variant,
        original_variants,
        auto_select_variant_return_value,
        expected_original_enable_calls,
        expected_variant_enable_calls,
        FakeConfig,
    ):
        m_ent = m_entitlement_factory.return_value
        m_ent.applicability_status.return_value = original_applicability_status
        m_ent.enable.return_value = (True, None)
        m_ent.is_variant = original_is_variant
        m_ent.variants = original_variants
        if auto_select_variant_return_value is not None:
            auto_select_variant_return_value[0].enable.return_value = (
                True,
                None,
            )
            m_auto_select_variant.return_value = (
                auto_select_variant_return_value
            )
        else:
            m_auto_select_variant.return_value = (m_ent, None)

        result = _enable(
            mock.MagicMock(), FakeConfig(), progress_object=mock.MagicMock()
        )

        assert expected_original_enable_calls == m_ent.enable.call_args_list
        if auto_select_variant_return_value is not None:
            assert (
                expected_variant_enable_calls
                == auto_select_variant_return_value[0].enable.call_args_list
            )
            assert len(result.warnings) == 1


class TestAutoSelectVariant:
    @pytest.mark.parametrize(
        [
            "variant_applicability_statuses",
            "variant_auto_selects",
            "default_variant",
            "expected_result",
        ],
        [
            # no variants
            (
                [],
                [],
                None,
                "original",
            ),
            # one variant, not applicable
            (
                [entitlements.ApplicabilityStatus.INAPPLICABLE],
                [False],
                None,
                "original",
            ),
            # one variant, applicable, but not autoselected
            (
                [entitlements.ApplicabilityStatus.APPLICABLE],
                [False],
                None,
                "original",
            ),
            # one variant, applicable, but not autoselected, with default
            (
                [entitlements.ApplicabilityStatus.APPLICABLE],
                [False],
                mock.MagicMock(),
                "default",
            ),
            # two variants, applicable, one autoselected, with default
            (
                [
                    entitlements.ApplicabilityStatus.APPLICABLE,
                    entitlements.ApplicabilityStatus.APPLICABLE,
                ],
                [False, True],
                mock.MagicMock(),
                1,
            ),
            # three variants, applicable, one autoselected, with default
            (
                [
                    entitlements.ApplicabilityStatus.INAPPLICABLE,
                    entitlements.ApplicabilityStatus.APPLICABLE,
                    entitlements.ApplicabilityStatus.APPLICABLE,
                ],
                [False, False, True],
                mock.MagicMock(),
                2,
            ),
            # three variants, applicable, two autoselected, takes first one
            (
                [
                    entitlements.ApplicabilityStatus.INAPPLICABLE,
                    entitlements.ApplicabilityStatus.APPLICABLE,
                    entitlements.ApplicabilityStatus.APPLICABLE,
                ],
                [False, True, True],
                mock.MagicMock(),
                1,
            ),
        ],
    )
    def test_auto_select_variant(
        self,
        variant_applicability_statuses,
        variant_auto_selects,
        default_variant,
        expected_result,
        FakeConfig,
    ):
        original = mock.MagicMock(name="original")
        original.default_variant = default_variant
        variants = []
        for i in range(len(variant_auto_selects)):
            v = mock.MagicMock(name="variant" + str(i))
            v.return_value.applicability_status.return_value = (
                variant_applicability_statuses[i],
                None,
            )
            v.return_value.variant_auto_select.return_value = (
                variant_auto_selects[i]
            )
            variants.append(v)

        if expected_result == "original":
            expected = (original, None)
        elif expected_result == "default":
            expected = (default_variant.return_value, mock.ANY)
        else:
            expected = (variants[expected_result].return_value, mock.ANY)

        assert expected == _auto_select_variant(
            FakeConfig(),
            mock.MagicMock(),
            original,
            variants,
            False,
        )



import json
import logging
import os
import pathlib
from collections import OrderedDict
from typing import Any, Dict, List, Union  # noqa: F401

from uaclient import defaults, secret_manager, system, util
from uaclient.config import UAConfig


class RegexRedactionFilter(logging.Filter):
    """A logging filter to redact confidential info"""

    def filter(self, record: logging.LogRecord):
        record.msg = util.redact_sensitive_logs(str(record.msg))
        return True


class KnownSecretRedactionFilter(logging.Filter):
    """A logging filter to redact confidential info"""

    def filter(self, record: logging.LogRecord):
        record.msg = secret_manager.secrets.redact_secrets(str(record.msg))
        return True


class JsonArrayFormatter(logging.Formatter):
    """Json Array Formatter for our logging mechanism
    Custom made for Pro logging needs
    """

    default_time_format = "%Y-%m-%dT%H:%M:%S"
    default_msec_format = "%s.%03d"
    required_fields = (
        "asctime",
        "levelname",
        "name",
        "funcName",
        "lineno",
        "message",
    )

    def format(self, record: logging.LogRecord) -> str:
        record.message = record.getMessage()
        record.asctime = self.formatTime(record)

        extra_message_dict = {}  # type: Dict[str, Any]
        if record.exc_info:
            extra_message_dict["exc_info"] = self.formatException(
                record.exc_info
            )
        if not extra_message_dict.get("exc_info") and record.exc_text:
            extra_message_dict["exc_info"] = record.exc_text
        if record.stack_info:
            extra_message_dict["stack_info"] = self.formatStack(
                record.stack_info
            )
        extra = record.__dict__.get("extra")
        if extra and isinstance(extra, dict):
            extra_message_dict.update(extra)

        # is ordered to maintain order of fields in log output
        local_log_record = OrderedDict()  # type: Dict[str, Any]
        # update the required fields in the order stated
        for field in self.required_fields:
            value = record.__dict__.get(field)
            local_log_record[field] = value

        local_log_record["extra"] = extra_message_dict
        return json.dumps(list(local_log_record.values()))


def get_user_or_root_log_file_path() -> str:
    """
    Gets the correct log_file path,
    adjusting for whether the user is root or not.
    """
    if util.we_are_currently_root():
        return UAConfig().log_file
    else:
        return get_user_log_file()


def get_user_log_file() -> str:
    """Gets the correct user log_file storage location"""
    return os.path.join(system.get_user_cache_dir(), "ubuntu-pro.log")


def get_all_user_log_files() -> List[str]:
    """Gets all the log files for the users in the system

    Returns a list of all user log files in their home directories.
    """
    user_directories = os.listdir("/home")
    log_files = []
    for user_directory in user_directories:
        user_path = os.path.join(
            "/home",
            user_directory,
            ".cache",
            defaults.USER_CACHE_SUBDIR,
            "ubuntu-pro.log",
        )
        if os.path.isfile(user_path):
            log_files.append(user_path)
    return log_files


def setup_journald_logging():
    logger = logging.getLogger("ubuntupro")
    logger.setLevel(logging.INFO)
    console_handler = logging.StreamHandler()
    console_handler.setFormatter(JsonArrayFormatter())
    console_handler.setLevel(logging.INFO)
    console_handler.addFilter(RegexRedactionFilter())
    console_handler.addFilter(KnownSecretRedactionFilter())
    logger.addHandler(console_handler)


def setup_cli_logging(log_level: Union[str, int], log_file: str):
    """Setup logging to log_file

    If run as non-root then log_file is replaced with a user-specific log file.
    """
    # support lower-case log_level config value
    if isinstance(log_level, str):
        log_level = log_level.upper()

    # if we are running as non-root, change log file
    if not util.we_are_currently_root():
        log_file = get_user_log_file()

    logger = logging.getLogger("ubuntupro")
    logger.setLevel(log_level)

    # Clear all handlers, so they are replaced for this logger
    logger.handlers = []

    # Setup file logging
    log_file_path = pathlib.Path(log_file)
    if not log_file_path.exists():
        log_file_path.parent.mkdir(parents=True, exist_ok=True)
        log_file_path.touch(mode=0o640)
    file_handler = logging.FileHandler(log_file)
    file_handler.setFormatter(JsonArrayFormatter())
    file_handler.setLevel(log_level)
    file_handler.addFilter(RegexRedactionFilter())
    file_handler.addFilter(KnownSecretRedactionFilter())

    logger.addHandler(file_handler)


def extra(**kwargs):
    """
    A helper for passing extra fields to log statements.
    Usage:
    LOG.info("message string", extra=log.extra(field=something_relevant))
    """
    return {"extra": kwargs}



import logging
import os
from subprocess import TimeoutExpired

from uaclient import exceptions, system, util
from uaclient.config import UAConfig
from uaclient.defaults import DEFAULT_DATA_DIR

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))

AUTO_ATTACH_STATUS_MOTD_FILE = os.path.join(
    DEFAULT_DATA_DIR, "messages", "motd-auto-attach-status"
)


def start():
    try:
        system.subp(
            ["systemctl", "start", "ubuntu-advantage.service"], timeout=2.0
        )
    except (exceptions.ProcessExecutionError, TimeoutExpired) as e:
        LOG.warning(e, exc_info=e)


def stop():
    try:
        system.subp(
            ["systemctl", "stop", "ubuntu-advantage.service"], timeout=2.0
        )
    except (exceptions.ProcessExecutionError, TimeoutExpired) as e:
        LOG.warning(e, exc_info=e)


def cleanup(cfg: UAConfig):
    from uaclient.daemon import retry_auto_attach

    retry_auto_attach.cleanup(cfg)



import logging
import os
import sys
import time

from uaclient import http, log, system
from uaclient.config import UAConfig
from uaclient.daemon import poll_for_pro_license, retry_auto_attach

LOG = logging.getLogger("ubuntupro.daemon")


# 10 seconds times 120 = 20 minutes
WAIT_FOR_CLOUD_CONFIG_SLEEP_TIME = 10
WAIT_FOR_CLOUD_CONFIG_POLL_TIMES = 120


def _wait_for_cloud_config():
    LOG.debug("waiting for cloud-config.service to finish")
    for i in range(WAIT_FOR_CLOUD_CONFIG_POLL_TIMES + 1):
        state = system.get_systemd_unit_active_state("cloud-config.service")
        ci_state = system.get_systemd_unit_active_state("cloud-init.service")
        LOG.debug("cloud-config.service state: %r", state)
        LOG.debug("cloud-init.service state: %r", ci_state)
        # if cloud-config.service is not yet activating but cloud-init is
        # running, wait for cloud-config to start
        if state == "activating" or (
            state == "inactive"
            and (ci_state == "activating" or ci_state == "active")
        ):
            if i < WAIT_FOR_CLOUD_CONFIG_POLL_TIMES:
                LOG.debug(
                    "cloud-config.service is activating. "
                    "waiting to check again."
                )
                time.sleep(WAIT_FOR_CLOUD_CONFIG_SLEEP_TIME)
            else:
                LOG.warning(
                    "cloud-config.service is still activating after "
                    "20 minutes. continuing anyway"
                )
                return
        else:
            LOG.debug("cloud-config.service is not activating. continuing")
            return


def main() -> int:
    log.setup_journald_logging()

    cfg = UAConfig()

    http.configure_web_proxy(cfg.http_proxy, cfg.https_proxy)

    LOG.info("daemon starting")

    _wait_for_cloud_config()

    LOG.debug("checking for condition files")
    is_correct_cloud = any(
        os.path.exists("/run/cloud-init/cloud-id-{}".format(cloud))
        for cloud in ("gce", "azure", "lxd")
    )
    if is_correct_cloud and not os.path.exists(
        retry_auto_attach.FLAG_FILE_PATH
    ):
        LOG.info("mode: poll for pro license")
        poll_for_pro_license.poll_for_pro_license(cfg)

    # not using elif because `poll_for_pro_license` may create the flag file

    if os.path.exists(retry_auto_attach.FLAG_FILE_PATH):
        LOG.info("mode: retry auto attach")
        retry_auto_attach.retry_auto_attach(cfg)

    LOG.info("daemon ending")
    return 0


if __name__ == "__main__":
    sys.exit(main())



import datetime
import glob
import logging
import os
import pathlib
import re
import stat
import subprocess  # nosec B404
import tempfile
import time
import uuid
from functools import lru_cache
from shutil import rmtree
from typing import Dict, List, NamedTuple, Optional, Sequence, Set, Tuple

from uaclient import defaults, exceptions, util

REBOOT_FILE_CHECK_PATH = "/var/run/reboot-required"
REBOOT_PKGS_FILE_PATH = "/var/run/reboot-required.pkgs"
ETC_MACHINE_ID = "/etc/machine-id"
DBUS_MACHINE_ID = "/var/lib/dbus/machine-id"
DISTRO_INFO_CSV = "/usr/share/distro-info/ubuntu.csv"

LOG = logging.getLogger(util.replace_top_level_logger_name(__name__))

# N.B. this relies on the version normalisation we perform in get_release_info
REGEX_OS_RELEASE_VERSION = (
    r"(?P<release>\d+\.\d+) (LTS\s*)?(\((?P<series>\w+))?.*"
)

RE_KERNEL_UNAME = (
    r"^"
    r"(?P<major>[\d]+)"
    r"[.-]"
    r"(?P<minor>[\d]+)"
    r"[.-]"
    r"(?P<patch>[\d]+)"
    r"-"
    r"(?P<abi>[\d]+)"
    r"-"
    r"(?P<flavor>[A-Za-z0-9_-]+)"
    r"$"
)

DistroInfo = NamedTuple(
    "DistroInfo",
    [
        ("eol", datetime.date),
        ("eol_esm", datetime.date),
        ("series", str),
        ("release", str),
        ("series_codename", str),
    ],
)

KernelInfo = NamedTuple(
    "KernelInfo",
    [
        ("uname_machine_arch", str),
        ("uname_release", str),
        ("build_date", Optional[datetime.datetime]),
        ("proc_version_signature_version", Optional[str]),
        ("major", Optional[int]),
        ("minor", Optional[int]),
        ("patch", Optional[int]),
        ("abi", Optional[str]),
        ("flavor", Optional[str]),
    ],
)

ReleaseInfo = NamedTuple(
    "ReleaseInfo",
    [
        ("distribution", str),
        ("release", str),
        ("series", str),
        ("pretty_version", str),
    ],
)

CpuInfo = NamedTuple(
    "CpuInfo",
    [
        ("cpuinfo_cpu", Optional[str]),
        ("cpuinfo_cpu_architecture", Optional[str]),
        ("cpuinfo_cpu_family", Optional[str]),
        ("cpuinfo_cpu_implementer", Optional[str]),
        ("cpuinfo_cpu_part", Optional[str]),
        ("cpuinfo_cpu_revision", Optional[str]),
        ("cpuinfo_cpu_variant", Optional[str]),
        ("cpuinfo_model", Optional[str]),
        ("cpuinfo_model_name", Optional[str]),
        ("cpuinfo_stepping", Optional[str]),
        ("cpuinfo_vendor_id", Optional[str]),
        ("sys_firmware_devicetree_base_model", Optional[str]),
        ("sysinfo_model", Optional[str]),
        ("sysinfo_type", Optional[str]),
    ],
)

RebootRequiredPkgs = NamedTuple(
    "RebootRequiredPkgs",
    [
        ("standard_packages", Optional[List[str]]),
        ("kernel_packages", Optional[List[str]]),
    ],
)


RE_KERNEL_EXTRACT_BUILD_DATE = r"(Mon|Tue|Wed|Thu|Fri|Sat|Sun).*"


def _get_kernel_changelog_timestamp(
    uname: os.uname_result,
) -> Optional[datetime.datetime]:
    if is_container():
        LOG.warning(
            "Not attempting to use timestamp of kernel "
            "changelog because we're in a container"
        )
        return None

    LOG.warning("Falling back to using timestamp of kernel changelog")

    try:
        stat_result = os.stat(
            "/usr/share/doc/linux-image-{}/changelog.Debian.gz".format(
                uname.release
            )
        )
        return datetime.datetime.fromtimestamp(
            stat_result.st_mtime, datetime.timezone.utc
        )
    except Exception:
        LOG.warning("Unable to stat kernel changelog")
        return None


def _get_kernel_build_date(
    uname: os.uname_result,
) -> Optional[datetime.datetime]:
    date_match = re.search(RE_KERNEL_EXTRACT_BUILD_DATE, uname.version)
    if date_match is None:
        LOG.warning("Unable to find build date in uname version")
        return _get_kernel_changelog_timestamp(uname)
    date_str = date_match.group(0)
    try:
        dt = datetime.datetime.strptime(date_str, "%a %b %d %H:%M:%S %Z %Y")
    except ValueError:
        LOG.warning("Unable to parse build date from uname version")
        return _get_kernel_changelog_timestamp(uname)
    if dt.tzinfo is None:
        # Give it a default timezone if it didn't get one from strptime
        # The Livepatch API requires a timezone
        dt = dt.replace(tzinfo=datetime.timezone.utc)
    return dt


@lru_cache(maxsize=None)
def get_kernel_info() -> KernelInfo:
    proc_version_signature_version = None
    try:
        proc_version_signature_full = load_file("/proc/version_signature")
        proc_version_signature_version = proc_version_signature_full.split()[1]
    except Exception:
        LOG.warning("failed to process /proc/version_signature.")

    uname = os.uname()
    uname_machine_arch = uname.machine.strip()
    build_date = _get_kernel_build_date(uname)

    uname_release = uname.release.strip()
    uname_match = re.match(RE_KERNEL_UNAME, uname_release)
    if uname_match is None:
        LOG.warning("Failed to parse kernel: %s", uname_release)
        return KernelInfo(
            uname_machine_arch=uname_machine_arch,
            uname_release=uname_release,
            build_date=build_date,
            proc_version_signature_version=proc_version_signature_version,
            major=None,
            minor=None,
            patch=None,
            abi=None,
            flavor=None,
        )
    else:
        return KernelInfo(
            uname_machine_arch=uname_machine_arch,
            uname_release=uname_release,
            build_date=build_date,
            proc_version_signature_version=proc_version_signature_version,
            major=int(uname_match.group("major")),
            minor=int(uname_match.group("minor")),
            patch=int(uname_match.group("patch")),
            abi=uname_match.group("abi"),
            flavor=uname_match.group("flavor"),
        )


# This only works if we are root (because of permissions to 'file' the kernels
# in /boot), but can't assert_root here due to circular imports.
def get_installed_ubuntu_kernels():
    # cursed circular import
    from uaclient.apt import get_installed_packages_names

    if not util.we_are_currently_root():
        raise RuntimeError(
            "get_installed_ubuntu_kernels needs to be executed as root"
        )

    linux_image = [
        package
        for package in get_installed_packages_names()
        if "linux-image-" in package
    ]
    vmlinux_kernel_files = [
        file
        for file in glob.glob("/boot/vmlinu[x|z]-*")
        if "Linux kernel" in subp(["file", file])[0]
    ]

    linux_image_versions = [
        package_name[len("linux-image-") :] for package_name in linux_image
    ]
    vmlinuz_versions = [
        file_name[len("/boot/vmlinu?-") :]
        for file_name in vmlinux_kernel_files
    ]

    return [
        version
        for version in vmlinuz_versions
        if version in linux_image_versions
    ]


@lru_cache(maxsize=None)
def get_dpkg_arch() -> str:
    out, _err = subp(["dpkg", "--print-architecture"])
    return out.strip()


@lru_cache(maxsize=None)
def get_virt_type() -> str:
    try:
        out, _ = subp(["systemd-detect-virt"])
        return out.strip()
    except exceptions.ProcessExecutionError:
        # The main known place where that will fail is in a docker/podman
        # container that doesn't have it installed. So we look for hints
        # of that situation to report it accurately.
        try:
            proc_1_cgroup = load_file("/proc/1/cgroup")
            if "docker" in proc_1_cgroup or "buildkit" in proc_1_cgroup:
                return "docker"
            elif "buildah" in proc_1_cgroup:
                return "podman"
            else:
                return ""
        except Exception:
            return ""


@lru_cache(maxsize=None)
def get_cpu_info() -> CpuInfo:
    CPUINFO_FILE = "/proc/cpuinfo"
    SYSINFO_FILE = "/proc/sysinfo"
    DEVICETREE_FILE = "/sys/firmware/devicetree/base/model"

    cpuinfo_values = {}
    sysinfo_values = {}
    devicetree_base_model = None

    try:
        cpuinfo_content = load_file(CPUINFO_FILE)
    except FileNotFoundError:
        LOG.info("could not open {}".format(CPUINFO_FILE))
    else:
        # There may be more than one cpu, we want the first.
        # Each cpu section is separated by a blank line.
        # The very first section may not be processor info, so we look for the
        # first section that starts with "processor"
        cpuinfo_sections = cpuinfo_content.split("\n\n")
        first_cpu_content = next(
            (s for s in cpuinfo_sections if s.startswith("processor")), ""
        )
        for line in first_cpu_content.split("\n"):
            if ":" in line:
                key, value = line.split(":", 1)
                cpuinfo_values[key.strip()] = value.strip()

    try:
        sysinfo_content = load_file(SYSINFO_FILE)
    except FileNotFoundError:
        LOG.info("could not open {}".format(SYSINFO_FILE))
    else:
        for line in sysinfo_content.split("\n"):
            if ":" in line:
                key, value = line.split(":", 1)
                sysinfo_values[key.strip()] = value.strip()

    try:
        devicetree_base_model = load_file(DEVICETREE_FILE)
    except FileNotFoundError:
        LOG.info("could not open {}".format(DEVICETREE_FILE))

    return CpuInfo(
        cpuinfo_cpu=cpuinfo_values.get("cpu"),
        cpuinfo_cpu_architecture=cpuinfo_values.get("CPU architecture"),
        cpuinfo_cpu_family=cpuinfo_values.get("cpu family"),
        cpuinfo_cpu_implementer=cpuinfo_values.get("CPU implementer"),
        cpuinfo_cpu_part=cpuinfo_values.get("CPU part"),
        cpuinfo_cpu_revision=cpuinfo_values.get("CPU revision"),
        cpuinfo_cpu_variant=cpuinfo_values.get("CPU variant"),
        cpuinfo_model=cpuinfo_values.get("model"),
        cpuinfo_model_name=cpuinfo_values.get("model name"),
        cpuinfo_stepping=cpuinfo_values.get("stepping"),
        cpuinfo_vendor_id=cpuinfo_values.get("vendor_id"),
        sys_firmware_devicetree_base_model=devicetree_base_model,
        sysinfo_model=sysinfo_values.get("Model"),
        sysinfo_type=sysinfo_values.get("Type"),
    )


@lru_cache(maxsize=None)
def get_machine_id(cfg) -> str:
    """
    Get system's unique machine-id or create our own in data_dir.
    We first check for the machine-id in machine-token.json before
    looking at the system file.
    """
    from uaclient.files import machine_token
    from uaclient.files.state_files import machine_id_file

    machine_token_file = machine_token.get_machine_token_file()
    if machine_token_file.machine_token:
        machine_id = machine_token_file.machine_token.get(
            "machineTokenInfo", {}
        ).get("machineId")
        if machine_id:
            return machine_id

    fallback_machine_id = machine_id_file.read()

    for path in [ETC_MACHINE_ID, DBUS_MACHINE_ID]:
        if os.path.exists(path):
            content = load_file(path).rstrip("\n")
            if content:
                return content

    if fallback_machine_id:
        return fallback_machine_id

    machine_id = str(uuid.uuid4())
    machine_id_file.write(machine_id)
    return machine_id


@lru_cache(maxsize=None)
def get_release_info() -> ReleaseInfo:
    os_release = _parse_os_release()
    distribution = os_release.get("NAME", "UNKNOWN")
    pretty_version = re.sub(r"\.\d LTS", " LTS", os_release.get("VERSION", ""))
    series = os_release.get("VERSION_CODENAME", "")
    release = os_release.get("VERSION_ID", "")

    if not series or not release:
        match = re.match(REGEX_OS_RELEASE_VERSION, pretty_version)
        if not match:
            raise exceptions.ParsingErrorOnOSReleaseFile(
                orig_ver=os_release.get("VERSION", ""), mod_ver=pretty_version
            )

        match_dict = match.groupdict()
        series = series or match_dict.get("series", "")
        if not series:
            raise exceptions.MissingSeriesOnOSReleaseFile(
                version=pretty_version
            )

        release = release or match_dict.get("release", "")

    return ReleaseInfo(
        distribution=distribution,
        release=release,
        series=series.lower(),
        pretty_version=pretty_version,
    )


@lru_cache(maxsize=None)
def is_lts(series: str) -> bool:
    out, _err = subp(["/usr/bin/ubuntu-distro-info", "--supported-esm"])
    return series in out


@lru_cache(maxsize=None)
def is_current_series_lts() -> bool:
    return is_lts(get_release_info().series)


@lru_cache(maxsize=None)
def is_supported(series: str) -> bool:
    out, _err = subp(["/usr/bin/ubuntu-distro-info", "--supported"])
    return series in out


@lru_cache(maxsize=None)
def is_active_esm(series: str) -> bool:
    """Return True when Ubuntu series supports ESM and is actively in ESM."""
    if not is_lts(series):
        return False
    out, _err = subp(
        ["/usr/bin/ubuntu-distro-info", "--series", series, "-yeol"]
    )
    return int(out) <= 0


@lru_cache(maxsize=None)
def is_current_series_active_esm() -> bool:
    return is_active_esm(get_release_info().series)


@lru_cache(maxsize=None)
def is_container(run_path: str = "/run") -> bool:
    """Checks to see if this code running in a container of some sort"""

    # We may mistake schroot environments for containers by just relying
    # in the other checks present in that function. To guarantee that
    # we do not identify a schroot as a container, we are explicitly
    # using the 'ischroot' command here.
    try:
        subp(["ischroot"])
        return False
    except exceptions.ProcessExecutionError:
        pass

    try:
        subp(["systemd-detect-virt", "--quiet", "--container"])
        return True
    except (IOError, OSError):
        pass

    for filename in ("container_type", "systemd/container"):
        path = os.path.join(run_path, filename)
        if os.path.exists(path):
            return True
    return False


@lru_cache(maxsize=None)
def is_desktop() -> bool:
    """Returns True if any package installed has "ubuntu-desktop" in the name.

    This includes ubuntu-desktop, ubuntu-desktop-minimal, kubuntu-desktop, etc.
    """
    from uaclient import apt

    for package in apt.get_installed_packages():
        if "ubuntu-desktop" in package.name:
            return True
    return False


@lru_cache(maxsize=None)
def _parse_os_release() -> Dict[str, str]:
    try:
        file_contents = load_file("/etc/os-release")
    except FileNotFoundError:
        file_contents = load_file("/usr/lib/os-release")
    data = {}
    for line in file_contents.splitlines():
        key, value = line.split("=", 1)
        if value:
            data[key] = value.strip().strip('"')
    return data


@lru_cache(maxsize=None)
def get_distro_info(series: str) -> DistroInfo:
    try:
        lines = load_file(DISTRO_INFO_CSV).splitlines()
    except FileNotFoundError:
        raise exceptions.MissingDistroInfoFile()
    for line in lines:
        values = line.split(",")
        if values[2] == series:
            if series == "xenial":
                eol_esm = "2026-04-23"
            else:
                eol_esm = values[7] if "LTS" in values[0] else values[5]
            return DistroInfo(
                release=values[0],
                series_codename=values[1],
                series=values[2],
                eol=datetime.datetime.strptime(values[5], "%Y-%m-%d").date(),
                eol_esm=datetime.datetime.strptime(eol_esm, "%Y-%m-%d").date(),
            )

    raise exceptions.MissingSeriesInDistroInfoFile(series=series)


def which(program: str) -> Optional[str]:
    """Find whether the provided program is executable in our PATH"""
    if os.path.sep in program:
        # if program had a '/' in it, then do not search PATH
        if is_exe(program):
            return program
    paths = [
        p.strip('"') for p in os.environ.get("PATH", "").split(os.pathsep)
    ]
    normalized_paths = [os.path.abspath(p) for p in paths]
    for path in normalized_paths:
        program_path = os.path.join(path, program)
        if is_exe(program_path):
            return program_path
    return None


def should_reboot(
    installed_pkgs: Optional[Set[str]] = None,
    installed_pkgs_regex: Optional[Set[str]] = None,
) -> bool:
    """Check if the system needs to be rebooted.

    :param installed_pkgs: If provided, verify if the any packages in
        the list are present on /var/run/reboot-required.pkgs. If that
        param is provided, we will only return true if we have the
        reboot-required marker file and any package in reboot-required.pkgs
        file. When both installed_pkgs and installed_pkgs_regex are
        provided, they act as an OR, so only one of the two lists must have
        a match to return True.
    :param installed_pkgs_regex: If provided, verify if the any regex in
        the list matches any line in /var/run/reboot-required.pkgs. If that
        param is provided, we will only return true if we have the
        reboot-required marker file and any match in reboot-required.pkgs
        file. When both installed_pkgs and installed_pkgs_regex are
        provided, they act as an OR, so only one of the two lists must have
        a match to return True.
    """

    # If the reboot marker file doesn't exist, we don't even
    # need to look at the installed_pkgs param
    if not os.path.exists(REBOOT_FILE_CHECK_PATH):
        return False

    # If there is no installed_pkgs to check, we will rely only
    # on the existence of the reboot marker file
    if installed_pkgs is None and installed_pkgs_regex is None:
        return True

    try:
        reboot_required_pkgs = set(
            load_file(REBOOT_PKGS_FILE_PATH).split("\n")
        )
    except FileNotFoundError:
        # If the file doesn't exist, we will default to the
        # reboot  marker file
        return True

    if installed_pkgs is not None:
        if len(installed_pkgs.intersection(reboot_required_pkgs)) != 0:
            return True

    if installed_pkgs_regex is not None:
        for pkg_name in reboot_required_pkgs:
            for pkg_regex in installed_pkgs_regex:
                if re.search(pkg_regex, pkg_name):
                    return True

    return False


def is_exe(path: str) -> bool:
    # return boolean indicating if path exists and is executable.
    return os.path.isfile(path) and os.access(path, os.X_OK)


def load_file(filename: str) -> str:
    """Read filename and decode content."""
    with open(filename, "rb") as stream:
        LOG.debug("Reading file: %s", filename)
        content = stream.read()
    try:
        return content.decode("utf-8")
    except UnicodeDecodeError:
        raise exceptions.InvalidFileEncodingError(
            file_name=filename, file_encoding="utf-8"
        )


def create_file(filename: str, mode: int = 0o644) -> None:
    LOG.debug("Creating file: %s", filename)
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    pathlib.Path(filename).touch()
    os.chmod(filename, mode)


def write_file(
    filename: str, content: str, mode: Optional[int] = None
) -> None:
    """Write content to the provided filename encoding it if necessary.

    We preserve the file ownership and permissions if the file is present
    and no mode argument is provided.

    @param filename: The full path of the file to write.
    @param content: The content to write to the file.
    @param mode: The filesystem mode to set on the file.
    """
    tmpf = None
    is_file_present = os.path.isfile(filename)
    if is_file_present:
        file_stat = pathlib.Path(filename).stat()
        f_mode = stat.S_IMODE(file_stat.st_mode)
        if mode is None:
            mode = f_mode

    elif mode is None:
        mode = 0o644
    try:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        tmpf = tempfile.NamedTemporaryFile(
            mode="wb", delete=False, dir=os.path.dirname(filename)
        )
        LOG.debug(
            "Writing file %s atomically via tempfile %s", filename, tmpf.name
        )
        tmpf.write(content.encode("utf-8"))
        tmpf.flush()
        tmpf.close()
        os.chmod(tmpf.name, mode)
        if is_file_present:
            os.chown(tmpf.name, file_stat.st_uid, file_stat.st_gid)
        os.rename(tmpf.name, filename)
    except Exception as e:
        if tmpf is not None:
            os.unlink(tmpf.name)
        raise e


def ensure_file_absent(file_path: str) -> None:
    """Remove a file if it exists, logging a message about removal."""
    try:
        os.unlink(file_path)
        LOG.debug("Removed file: %s", file_path)
    except FileNotFoundError:
        LOG.debug("Tried to remove %s but file does not exist", file_path)


def _subp(
    args: Sequence[str],
    rcs: Optional[List[int]] = None,
    capture: bool = False,
    timeout: Optional[float] = None,
    override_env_vars: Optional[Dict[str, str]] = None,
    pipe_stdouterr: bool = True,
) -> Tuple[str, str]:
    """Run a command and return a tuple of decoded stdout, stderr.

    @param args: A list of arguments to feed to subprocess.Popen
    @param rcs: A list of allowed return_codes. If returncode not in rcs
        raise a ProcessExecutionError.
    @param capture: Boolean set True to log the command and response.
    @param timeout: Optional float indicating number of seconds to wait for
        subp to return.
    @param override_env_vars: Optional dictionary of environment variables.
        If None, the current os.environ is used for the subprocess.
        If defined, these env vars get merged with the current process'
        os.environ for the subprocess, overriding any values that already
        existed in os.environ.

    @return: Tuple of utf-8 decoded stdout, stderr
    @raises ProcessExecutionError on invalid command or returncode not in rcs.
    @raises subprocess.TimeoutError when timeout specified and the command
        exceeds that number of seconds.
    """
    bytes_args = [
        x if isinstance(x, bytes) else x.encode("utf-8") for x in args
    ]

    stdout = None
    stderr = None
    set_lang = {}

    if pipe_stdouterr:
        stdout = subprocess.PIPE
        stderr = subprocess.PIPE
        # Set LANG to avoid non-utf8 when we pipe the handlers
        set_lang = {"LANG": "C.UTF-8", "LC_ALL": "C.UTF-8"}

    if override_env_vars is None:
        override_env_vars = {}
    merged_env = {**os.environ, **set_lang, **override_env_vars}

    if rcs is None:
        rcs = [0]
    redacted_cmd = util.redact_sensitive_logs(" ".join(args))
    try:
        proc = subprocess.Popen(  # nosec B603
            bytes_args,
            stdout=stdout,
            stderr=stderr,
            env=merged_env,
        )
        (out, err) = proc.communicate(timeout=timeout)
    except OSError:
        try:
            out_result = out.decode("utf-8", errors="ignore") if out else ""
            err_result = err.decode("utf-8", errors="ignore") if err else ""
            raise exceptions.ProcessExecutionError(
                cmd=redacted_cmd,
                exit_code=proc.returncode,
                stdout=out_result,
                stderr=err_result,
            )
        except UnboundLocalError:
            raise exceptions.ProcessExecutionError(cmd=redacted_cmd)

    out_result = out.decode("utf-8", errors="ignore") if out else ""
    err_result = err.decode("utf-8", errors="ignore") if err else ""
    if proc.returncode not in rcs:
        raise exceptions.ProcessExecutionError(
            cmd=redacted_cmd,
            exit_code=proc.returncode,
            stdout=out_result,
            stderr=err_result,
        )
    if capture:
        LOG.debug(
            "Ran cmd: %s, rc: %s stderr: %s",
            redacted_cmd,
            proc.returncode,
            err,
        )
    return out_result, err_result


def subp(
    args: Sequence[str],
    rcs: Optional[List[int]] = None,
    capture: bool = False,
    timeout: Optional[float] = None,
    retry_sleeps: Optional[List[float]] = None,
    override_env_vars: Optional[Dict[str, str]] = None,
    pipe_stdouterr: bool = True,
) -> Tuple[str, str]:
    """Run a command and return a tuple of decoded stdout, stderr.

     @param subp: A list of arguments to feed to subprocess.Popen
     @param rcs: A list of allowed return_codes. If returncode not in rcs
         raise a ProcessExecutionError.
     @param capture: Boolean set True to log the command and response.
     @param timeout: Optional float indicating number of seconds to wait for a
         subp call to return.
     @param retry_sleeps: Optional list of sleep lengths to apply between
        retries. Specifying a list of [0.5, 1] instructs subp to retry twice
        on failure; sleeping half a second before the first retry and 1 second
        before the next retry.
     @param override_env_vars: Optional dictionary of environment variables.
        If None, the current os.environ is used for the subprocess.
        If defined, these env vars get merged with the current process'
        os.environ for the subprocess, overriding any values that already
        existed in os.environ.

    @return: Tuple of utf-8 decoded stdout, stderr
    @raises ProcessExecutionError on invalid command or returncode not in rcs.
    @raises subprocess.TimeoutError when timeout specified and the command
        exceeds that number of seconds.
    """
    retry_sleeps = retry_sleeps.copy() if retry_sleeps is not None else None
    while True:
        try:
            out, err = _subp(
                args,
                rcs,
                capture,
                timeout,
                override_env_vars=override_env_vars,
                pipe_stdouterr=pipe_stdouterr,
            )
            break
        except exceptions.ProcessExecutionError as e:
            if capture:
                LOG.debug(str(e))
                LOG.warning("Stderr: %s\nStdout: %s", e.stderr, e.stdout)
            if not retry_sleeps:
                raise
            LOG.debug(str(e))
            LOG.debug("Retrying %d more times.", len(retry_sleeps))
            time.sleep(retry_sleeps.pop(0))
    return out, err


def ensure_folder_absent(folder_path: str) -> None:
    try:
        rmtree(folder_path)
        LOG.debug("Removed folder: %s", folder_path)
    except FileNotFoundError:
        LOG.debug("Tried to remove %s but folder does not exist", folder_path)


def is_systemd_unit_active(service_name: str) -> bool:
    """
    Get if the systemd job is active in the system. Note that any status
    different from "active" will make this function return False.
    Additionally, if the system doesn't exist we will also return False
    here.

    @param service_name: Name of the systemd job to look at

    @return: A Boolean specifying if the job is active or not
    """
    try:
        subp(["systemctl", "is-active", "--quiet", service_name])
    except exceptions.ProcessExecutionError:
        return False
    return True


def get_systemd_unit_active_state(service_name: str) -> Optional[str]:
    try:
        out, _ = subp(
            [
                "systemctl",
                "show",
                "--property=ActiveState",
                "--no-pager",
                service_name,
            ]
        )
        if out and out.startswith("ActiveState="):
            return out.split("=")[1].strip()
        else:
            LOG.warning(
                "Couldn't find ActiveState in systemctl show output for %s",
                service_name,
            )
    except exceptions.ProcessExecutionError as e:
        LOG.warning(
            "Failed to get ActiveState for systemd unit %s",
            service_name,
            exc_info=e,
        )
    return None


def get_user_cache_dir() -> str:
    if util.we_are_currently_root():
        return defaults.UAC_RUN_PATH

    xdg_cache_home = os.environ.get("XDG_CACHE_HOME")
    if xdg_cache_home:
        return os.path.join(xdg_cache_home, defaults.USER_CACHE_SUBDIR)

    return os.path.join(
        os.path.expanduser("~"), ".cache", defaults.USER_CACHE_SUBDIR
    )


def get_reboot_required_pkgs() -> Optional[RebootRequiredPkgs]:
    try:
        pkg_list_str = load_file(REBOOT_PKGS_FILE_PATH)
    except FileNotFoundError:
        return None

    standard_packages = []
    kernel_packages = []
    kernel_regex = "^(linux-image|linux-base).*"

    for pkg in pkg_list_str.split():
        if re.match(kernel_regex, pkg):
            kernel_packages.append(pkg)
        else:
            standard_packages.append(pkg)

    return RebootRequiredPkgs(
        standard_packages=sorted(standard_packages),
        kernel_packages=sorted(kernel_packages),
    )



import mock
import pytest

from uaclient import exceptions, lock, messages
from uaclient.cli.refresh import refresh_command
from uaclient.files.notices import Notice

M_PATH = "uaclient.cli.refresh."


class TestActionRefresh:
    @mock.patch("uaclient.util.we_are_currently_root", return_value=False)
    def test_non_root_users_are_rejected(
        self, we_are_currently_root, FakeConfig, fake_machine_token_file
    ):
        """Check that a UID != 0 will receive a message and exit non-zero"""

        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        with pytest.raises(exceptions.NonRootUserError):
            refresh_command.action(mock.MagicMock(), cfg=cfg)

    @pytest.mark.parametrize(
        "target, expect_unattached_error",
        [(None, True), ("contract", True), ("config", False)],
    )
    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    def test_not_attached_errors(
        self,
        _m_check_lock_info,
        target,
        expect_unattached_error,
        FakeConfig,
    ):
        """Check that an unattached machine emits message and exits 1"""
        cfg = FakeConfig()

        cfg.user_config.update_messaging_timer = 0
        cfg.user_config.metering_timer = 0
        cfg.process_config = lambda *a, **kw: None

        with mock.patch.object(lock, "lock_data_file"):
            if expect_unattached_error:
                with pytest.raises(exceptions.UnattachedError):
                    refresh_command.action(
                        mock.MagicMock(target=target), cfg=cfg
                    )
            else:
                refresh_command.action(mock.MagicMock(target=target), cfg=cfg)

    @mock.patch("uaclient.lock.check_lock_info")
    @mock.patch("time.sleep")
    @mock.patch("uaclient.system.subp")
    def test_lock_file_exists(
        self,
        m_subp,
        m_sleep,
        m_check_lock_info,
        FakeConfig,
        fake_machine_token_file,
    ):
        """Check inability to refresh if operation holds lock file."""
        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        m_check_lock_info.return_value = (123, "pro disable")
        with pytest.raises(exceptions.LockHeldError) as err:
            refresh_command.action(mock.MagicMock(), cfg=cfg)
        assert 12 == m_check_lock_info.call_count
        expected_msg = messages.E_LOCK_HELD_ERROR.format(
            lock_request="pro refresh", lock_holder="pro disable", pid=123
        )
        assert expected_msg.msg == err.value.msg

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("logging.exception")
    @mock.patch("uaclient.contract.refresh")
    @mock.patch("uaclient.files.notices.NoticesManager.remove")
    def test_refresh_contract_error_on_failure_to_update_contract(
        self,
        m_remove_notice,
        refresh,
        logging_error,
        m_check_log_info,
        FakeConfig,
        fake_machine_token_file,
    ):
        """On failure in request_updates_contract emit an error."""
        refresh.side_effect = exceptions.ConnectivityError(
            mock.MagicMock(), "url"
        )

        cfg = FakeConfig()
        fake_machine_token_file.attached = True

        with pytest.raises(exceptions.UbuntuProError) as excinfo:
            with mock.patch.object(lock, "lock_data_file"):
                refresh_command.action(
                    mock.MagicMock(target="contract"), cfg=cfg
                )

        assert messages.E_REFRESH_CONTRACT_FAILURE.msg == excinfo.value.msg
        assert [
            mock.call("", messages.NOTICE_REFRESH_CONTRACT_WARNING)
        ] != m_remove_notice.call_args_list

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("uaclient.contract.refresh")
    @mock.patch("uaclient.files.notices.NoticesManager.remove")
    def test_refresh_contract_happy_path(
        self,
        m_remove_notice,
        refresh,
        _m_check_lock_info,
        capsys,
        FakeConfig,
        fake_machine_token_file,
    ):
        """On success from request_updates_contract root user can refresh."""
        refresh.return_value = True

        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        with mock.patch.object(lock, "lock_data_file"):
            ret = refresh_command.action(
                mock.MagicMock(target="contract"), cfg=cfg
            )

        assert 0 == ret
        assert messages.REFRESH_CONTRACT_SUCCESS in capsys.readouterr()[0]
        assert [mock.call(cfg)] == refresh.call_args_list
        assert [
            mock.call(Notice.OPERATION_IN_PROGRESS),
        ] == m_remove_notice.call_args_list

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch(M_PATH + "update_motd_messages")
    def test_refresh_messages_error(
        self, m_update_motd, _m_check_lock_info, FakeConfig
    ):
        """On failure in update_motd_messages emit an error."""
        m_update_motd.side_effect = Exception("test")

        with pytest.raises(exceptions.UbuntuProError) as excinfo:
            with mock.patch.object(lock, "lock_data_file"):
                refresh_command.action(
                    mock.MagicMock(target="messages"), cfg=FakeConfig()
                )

        assert messages.E_REFRESH_MESSAGES_FAILURE.msg == excinfo.value.msg

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("uaclient.apt_news.update_apt_news")
    @mock.patch("uaclient.timer.update_messaging.exists", return_value=True)
    @mock.patch("uaclient.timer.update_messaging.LOG.exception")
    @mock.patch("uaclient.system.subp")
    @mock.patch(M_PATH + "update_motd_messages")
    def test_refresh_messages_doesnt_fail_if_update_notifier_does(
        self,
        m_update_motd,
        m_subp,
        log_exception,
        _m_path,
        _m_update_apt_news,
        _m_check_lock_info,
        capsys,
        FakeConfig,
    ):
        subp_exc = Exception("test")
        m_subp.side_effect = [subp_exc, ""]

        with mock.patch.object(lock, "lock_data_file"):
            ret = refresh_command.action(
                mock.MagicMock(target="messages"), cfg=FakeConfig()
            )

        assert 0 == ret
        assert 1 == log_exception.call_count
        assert [mock.call(subp_exc)] == log_exception.call_args_list
        assert messages.REFRESH_MESSAGES_SUCCESS in capsys.readouterr()[0]

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("uaclient.apt_news.update_apt_news")
    @mock.patch(M_PATH + "refresh_motd")
    @mock.patch(M_PATH + "update_motd_messages")
    def test_refresh_messages_happy_path(
        self,
        m_update_motd,
        m_refresh_motd,
        m_update_apt_news,
        _m_check_lock_info,
        capsys,
        FakeConfig,
    ):
        """On success from request_updates_contract root user can refresh."""
        cfg = FakeConfig()
        with mock.patch.object(lock, "lock_data_file"):
            ret = refresh_command.action(
                mock.MagicMock(target="messages"), cfg=cfg
            )

        assert 0 == ret
        assert messages.REFRESH_MESSAGES_SUCCESS in capsys.readouterr()[0]
        assert [mock.call(cfg)] == m_update_motd.call_args_list
        assert [mock.call()] == m_refresh_motd.call_args_list
        assert 1 == m_update_motd.call_count
        assert 1 == m_refresh_motd.call_count
        assert 1 == m_update_apt_news.call_count

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("logging.exception")
    @mock.patch(
        "uaclient.config.UAConfig.process_config", side_effect=RuntimeError()
    )
    def test_refresh_config_error_on_failure_to_process_config(
        self,
        _m_process_config,
        _m_logging_error,
        _m_check_lock_info,
        FakeConfig,
        fake_machine_token_file,
    ):
        """On failure in process_config emit an error."""

        cfg = FakeConfig()
        fake_machine_token_file.attached = True

        with pytest.raises(exceptions.UbuntuProError) as excinfo:
            with mock.patch.object(lock, "lock_data_file"):
                refresh_command.action(
                    mock.MagicMock(target="config"), cfg=cfg
                )

        assert messages.E_REFRESH_CONFIG_FAILURE.msg == excinfo.value.msg

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("uaclient.config.UAConfig.process_config")
    def test_refresh_config_happy_path(
        self,
        m_process_config,
        m_check_lock_info,
        capsys,
        FakeConfig,
        fake_machine_token_file,
    ):
        """On success from process_config root user gets success message."""

        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        with mock.patch.object(lock, "lock_data_file"):
            ret = refresh_command.action(
                mock.MagicMock(target="config"), cfg=cfg
            )

        assert 0 == ret
        assert messages.REFRESH_CONFIG_SUCCESS in capsys.readouterr()[0]
        assert [mock.call()] == m_process_config.call_args_list

    @mock.patch("uaclient.lock.check_lock_info", return_value=(-1, ""))
    @mock.patch("uaclient.apt_news.update_apt_news")
    @mock.patch(M_PATH + "refresh_motd")
    @mock.patch(M_PATH + "update_motd_messages")
    @mock.patch("uaclient.contract.refresh")
    @mock.patch("uaclient.config.UAConfig.process_config")
    @mock.patch("uaclient.files.notices.NoticesManager.remove")
    def test_refresh_all_happy_path(
        self,
        m_remove_notice,
        m_process_config,
        m_refresh,
        m_update_motd,
        m_refresh_motd,
        m_update_apt_news,
        _m_check_lock_info,
        capsys,
        FakeConfig,
        fake_machine_token_file,
    ):
        """On success from process_config root user gets success message."""

        cfg = FakeConfig()
        fake_machine_token_file.attached = True
        with mock.patch.object(lock, "lock_data_file"):
            ret = refresh_command.action(mock.MagicMock(target=None), cfg=cfg)

        out, err = capsys.readouterr()

        assert 0 == ret
        assert messages.REFRESH_CONFIG_SUCCESS in out
        assert messages.REFRESH_CONTRACT_SUCCESS in out
        assert messages.REFRESH_MESSAGES_SUCCESS in out
        assert [mock.call()] == m_process_config.call_args_list
        assert [mock.call(cfg)] == m_refresh.call_args_list
        assert [
            mock.call(Notice.OPERATION_IN_PROGRESS),
        ] == m_remove_notice.call_args_list
        assert 1 == m_update_motd.call_count
        assert 1 == m_refresh_motd.call_count
        assert 1 == m_update_apt_news.call_count



from typing import List


class SecretManager:
    def __init__(self):
        self._secrets = []

    def add_secret(self, secret: str) -> None:
        if secret:  # Add only non-empty secrets
            self._secrets.append(secret)

    @property
    def secrets(self) -> List[str]:
        return self._secrets

    def clear_secrets(self) -> None:
        self._secrets.clear()

    def redact_secrets(self, log_record: str) -> str:
        redacted_record = log_record
        for secret in self._secrets:
            redacted_record = redacted_record.replace(secret, "<REDACTED>")
        return redacted_record


secrets = SecretManager()



from typing import List, Optional, Tuple

from uaclient import actions, contract, event_logger, lock, messages, util
from uaclient.api import exceptions
from uaclient.api.api import APIEndpoint
from uaclient.api.data_types import AdditionalInfo
from uaclient.api.u.pro.status.is_attached.v1 import _is_attached
from uaclient.clouds import identity
from uaclient.config import UAConfig
from uaclient.data_types import DataObject, Field, StringDataValue, data_list
from uaclient.entitlements import order_entitlements_for_enabling
from uaclient.entitlements.entitlement_status import CanEnableFailure
from uaclient.files import machine_token

event = event_logger.get_event_logger()


class FullAutoAttachOptions(DataObject):
    fields = [
        Field(
            "enable",
            data_list(StringDataValue),
            False,
            doc="Optional list of services to enable after auto-attaching.",
        ),
        Field(
            "enable_beta",
            data_list(StringDataValue),
            False,
            doc=(
                "Optional list of beta services to enable after"
                " auto-attaching."
            ),
        ),
        Field(
            "cloud_override",
            StringDataValue,
            False,
            doc=(
                "Ignore the result of ``cloud-id`` and act as if running on"
                " this cloud."
            ),
        ),
    ]

    def __init__(
        self,
        enable: Optional[List[str]] = None,
        enable_beta: Optional[List[str]] = None,
        cloud_override: Optional[str] = None,
    ):
        self.enable = enable
        self.enable_beta = enable_beta
        self.cloud_override = cloud_override


class FullAutoAttachResult(DataObject, AdditionalInfo):
    pass


def _enable_services_by_name(
    cfg: UAConfig, services: List[str]
) -> List[Tuple[str, messages.NamedMessage]]:
    failed_services = []
    for name in order_entitlements_for_enabling(cfg, services):
        try:
            ent_ret, reason = actions.enable_entitlement_by_name(cfg, name)
        except exceptions.UbuntuProError as e:
            failed_services.append(
                (name, messages.NamedMessage(e.msg_code or "unknown", e.msg))
            )
            continue
        if not ent_ret:
            if (
                reason is not None
                and isinstance(reason, CanEnableFailure)
                and reason.message is not None
            ):
                failed_services.append((name, reason.message))
            else:
                failed_services.append(
                    (
                        name,
                        messages.NamedMessage("unknown", "failed to enable"),
                    )
                )
    return failed_services


def full_auto_attach(options: FullAutoAttachOptions) -> FullAutoAttachResult:
    return _full_auto_attach(options, UAConfig())


def _full_auto_attach(
    options: FullAutoAttachOptions,
    cfg: UAConfig,
    *,
    mode: event_logger.EventLoggerMode = event_logger.EventLoggerMode.JSON
) -> FullAutoAttachResult:
    """
    This endpoint runs the whole auto-attach process on the system.
    """
    try:
        with lock.RetryLock(
            lock_holder="pro.api.u.pro.attach.auto.full_auto_attach.v1",
        ):
            ret = _full_auto_attach_in_lock(options, cfg, mode=mode)
    except Exception as e:
        lock.clear_lock_file_if_present()
        raise e
    return ret


def _full_auto_attach_in_lock(
    options: FullAutoAttachOptions,
    cfg: UAConfig,
    mode: event_logger.EventLoggerMode,
) -> FullAutoAttachResult:
    event.set_event_mode(mode)
    machine_token_file = machine_token.get_machine_token_file(cfg)

    if _is_attached(cfg).is_attached:
        raise exceptions.AlreadyAttachedError(
            account_name=machine_token_file.account.get("name", "")
        )

    if util.is_config_value_true(
        config=cfg.cfg, path_to_value="features.disable_auto_attach"
    ):
        raise exceptions.AutoAttachDisabledError()

    instance = identity.cloud_instance_factory(
        cloud_override=options.cloud_override
    )
    enable_default_services = (
        options.enable is None and options.enable_beta is None
    )
    actions.auto_attach(cfg, instance, allow_enable=enable_default_services)

    failed = []
    if options.enable is not None:
        failed += _enable_services_by_name(cfg, options.enable)
    if options.enable_beta is not None:
        failed += _enable_services_by_name(cfg, options.enable_beta)

    contract_client = contract.UAContractClient(cfg)
    contract_client.update_activity_token()

    if len(failed) > 0:
        raise exceptions.EntitlementsNotEnabledError(failed)

    return FullAutoAttachResult()


endpoint = APIEndpoint(
    version="v1",
    name="FullAutoAttach",
    fn=_full_auto_attach,
    options_cls=FullAutoAttachOptions,
)

_doc = {
    "introduced_in": "27.11",
    "requires_network": True,
    "extra_args_content": """
.. note::

    If none of the lists are set, the services will be enabled based on the
    contract definitions.
""",
    "example_python": """
from uaclient.api.u.pro.attach.auto.full_auto_attach.v1 import full_auto_attach, FullAutoAttachOptions

options = FullAutoAttachOptions(enable=["<service1>", "<service2>"], enable_beta=["<beta_service3>"])
result = full_auto_attach(options)
""",  # noqa: E501
    "result_class": FullAutoAttachResult,
    "exceptions": [
        (
            exceptions.AlreadyAttachedError,
            (
                "Raised if running on a machine which is already attached to a"
                " Pro subscription."
            ),
        ),
        (
            exceptions.AutoAttachDisabledError,
            "Raised if ``disable_auto_attach: true`` in ``uaclient.conf``.",
        ),
        (
            exceptions.ConnectivityError,
            (
                "Raised if it is not possible to connect to the contracts"
                " service."
            ),
        ),
        (
            exceptions.ContractAPIError,
            (
                "Raised if there is an unexpected error in the contracts"
                " service interaction."
            ),
        ),
        (
            exceptions.EntitlementsNotEnabledError,
            (
                "Raised if the Client fails to enable any of the entitlements"
                " (whether present in any of the lists or listed in the"
                " contract)."
            ),
        ),
        (
            exceptions.LockHeldError,
            (
                "Raised if another Client process is holding the lock on the"
                " machine."
            ),
        ),
        (
            exceptions.NonAutoAttachImageError,
            (
                "Raised if the cloud where the system is running does not"
                " support auto-attach."
            ),
        ),
    ],
    "example_cli": 'pro api u.pro.attach.auto.full_auto_attach.v1 --data {"enable": ["esm-infra", "esm-apps"]}',  # noqa: E501
    "example_json": """
{}
""",
}
